# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from threading import Semaphore
from threading import Thread
from typing import Optional

from mkdocs.config.defaults import MkDocsConfig

# noinspection PyProtectedMember
from mkdocs_publisher._shared import file_utils
from mkdocs_publisher.minifier.config import MinifierConfig
from mkdocs_publisher.minifier.config import _MinifierCommonConfig

log = logging.getLogger("mkdocs.plugins.publisher.minifier.base")


@dataclass
class CachedFile:
    original_file_hash: Optional[str] = field(default="")
    original_file_path: Path = field(default_factory=lambda: Path(""))
    cached_file_name: Path = field(default_factory=lambda: Path(""))

    def __init__(
        self,
        original_file_hash: str = "",
        original_file_path: str = "",
        cached_file_name: str = "",
    ):
        self.original_file_hash = original_file_hash
        self.original_file_path = Path(original_file_path)
        self.cached_file_name = Path(cached_file_name)

    def based_on(self, file: Path, directory: Path):
        self.original_file_path = file
        self.original_file_hash = file_utils.calculate_file_hash(file=directory / file)
        self.cached_file_name = file_utils.get_hashed_file_name(file=file)

    @property
    def as_dict(self) -> dict:
        return asdict(self)


class BaseMinifier:
    def __init__(
        self,
        plugin_config: MinifierConfig,
        mkdocs_config: MkDocsConfig,
        cached_files: dict[str, CachedFile],
    ):
        self._plugin_config: MinifierConfig = plugin_config
        self._mkdocs_config: MkDocsConfig = mkdocs_config
        self._minify_options: Optional[_MinifierCommonConfig] = None
        self._cached_files: dict[str, CachedFile] = cached_files
        self._cache_enabled: bool = self._plugin_config.cache_enabled
        self._exclude: list = self._plugin_config.exclude[:]

    def minifier(self, cached_file: CachedFile) -> Optional[CachedFile]:
        raise NotImplementedError

    def __call__(self):
        minifier_name = self.__class__.__name__.replace("Minifier", "").upper()
        self._cache_enabled = self._minify_options.cache_enabled if self._cache_enabled else False
        self._exclude.extend(self._minify_options.exclude)

        log.info(
            f"Minifying {minifier_name} files ({'with' if self._cache_enabled else 'no'} cache)"
        )
        log.info(f"Excluded {minifier_name} files patterns: {self._exclude}")

        semaphore = Semaphore(self._plugin_config.threads)
        threads = []

        for file in file_utils.list_files(
            directory=Path(self._mkdocs_config.site_dir),
            extensions=self._minify_options.extensions,
            exclude=self._exclude,
        ):
            log.debug(f"Minifying: {file}")
            semaphore.acquire()
            thread = Thread(
                target=self._minify_with_cache,
                kwargs={
                    "file": file,
                    "cache_enabled": self._cache_enabled,
                    "semaphore": semaphore,
                },
            )
            thread.start()
            threads.append(thread)

        # Wait for all threads to be finished
        for thread in threads:
            thread.join()

    def _is_new_smaller(self, cached_file: CachedFile) -> Optional[CachedFile]:
        old_file = self._mkdocs_config.site_dir / cached_file.original_file_path
        new_file = self._plugin_config.cache_dir / cached_file.cached_file_name
        new_file_size = new_file.stat().st_size
        old_file_size = old_file.stat().st_size
        log.debug(
            f"Minified: '{old_file.relative_to(self._mkdocs_config.site_dir)}' "
            f"(size: {old_file_size} -> {new_file_size})"
        )
        log.debug(
            f"{old_file.relative_to(self._mkdocs_config.site_dir)} -> "
            f"{new_file.relative_to(self._plugin_config.cache_dir)} "
        )
        if new_file_size < old_file_size:
            return cached_file

        log.debug(
            f"Minified file larger than original: "
            f"{cached_file.original_file_path} (removing cached file)"
        )
        new_file.unlink()
        return None

    def _copy_cached_file(self, cached_file: CachedFile):
        original_file = self._mkdocs_config.site_dir / cached_file.original_file_path
        cache_file = self._plugin_config.cache_dir / cached_file.cached_file_name
        try:
            original_file.write_bytes(cache_file.read_bytes())
        except FileNotFoundError as e:
            log.warning(e)

    def _minify_with_cache(self, file: Path, cache_enabled: bool, semaphore: Semaphore):
        recreate_file = False
        if cache_enabled and str(file) in self._cached_files:
            log.debug(f"{file} is in cache")
            file_hash = file_utils.calculate_file_hash(file=(self._mkdocs_config.site_dir / file))
            cached_file = self._cached_files[str(file)]
            if file_hash is not None and cached_file.original_file_hash == file_hash:
                log.debug(f"{file} hash is equal to one in cache (file: {file_hash})")
                cached_file_name = self._plugin_config.cache_dir / cached_file.cached_file_name
                if cached_file_name.exists():
                    self._copy_cached_file(cached_file=cached_file)
                else:
                    recreate_file = True
            else:
                log.debug(
                    f"{file} hash is not equal to one in cache "
                    f"(file: {file_hash} | cache: {cached_file.original_file_hash})"
                )
                recreate_file = True
        else:
            log.debug(f"{file} is not in cache (rebuilding")

            cached_file = CachedFile()
            cached_file.based_on(file=file, directory=Path(self._mkdocs_config.site_dir))
            recreate_file = True

        if recreate_file:
            log.debug(f"Minifying file: {file} (putting into cache)")
            cached_file = self.minifier(cached_file=cached_file)
            if cached_file:
                cached_file = self._is_new_smaller(cached_file=cached_file)
            if cached_file:
                self._cached_files[str(cached_file.original_file_path)] = cached_file
                self._copy_cached_file(cached_file=cached_file)
        else:
            log.debug(f"{file} is already minified (retrieving from cache)")
        semaphore.release()



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mkdocs.config import config_options as option
from mkdocs.config.base import Config


class _BlogTranslationConfig(Config):
    teaser_link_text = option.Optional(option.Type(str))
    blog_page_title = option.Optional(option.Type(str))
    blog_navigation_name = option.Optional(option.Type(str))
    recent_blog_posts_navigation_name = option.Optional(option.Type(str))
    archive_page_title = option.Optional(option.Type(str))
    archive_navigation_name = option.Optional(option.Type(str))
    categories_page_title = option.Optional(option.Type(str))
    categories_navigation_name = option.Optional(option.Type(str))
    tags_page_title = option.Optional(option.Type(str))
    tags_navigation_name = option.Optional(option.Type(str))
    newer_posts = option.Optional(option.Type(str))
    older_posts = option.Optional(option.Type(str))


class _BlogCommentsConfig(Config):
    enabled = option.Type(bool, default=False)
    key_name = option.Type(str, default="comments")


class BlogPluginConfig(Config):
    # General settings
    lang = option.Choice(["en", "pl"], default="en")  # TODO: auto update based on files
    teaser_marker = option.Type(str, default="<!-- more -->")
    searchable_non_posts = option.Type(bool, default=False)
    posts_per_page = option.Type(int, default=5)
    slug = option.Type(str, default="blog")

    # Directories
    temp_dir = option.Type(str, default=".pub_blog_temp")
    blog_dir = option.Type(str, default="blog")
    archive_subdir = option.Type(str, default="archive")
    categories_subdir = option.Type(str, default="categories")
    tags_subdir = option.Type(str, default="tags")

    comments: _BlogCommentsConfig = option.SubConfig(_BlogCommentsConfig)  # type: ignore
    # Values that are in lang files and can be overriden
    translation: _BlogTranslationConfig = option.SubConfig(_BlogTranslationConfig)  # type: ignore



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from typing import Optional

from mkdocs_publisher.obsidian.config import _ObsidianCalloutsConfig

log = logging.getLogger("mkdocs.plugins.publisher.obsidian.callouts")

CALLOUT_BLOCK = re.compile(r"^ *(( ?>)+) *\[!([^]]*)]([\-+]?) *(.*)?")
CALLOUT_BLOCK_FOLLOW = re.compile(r"^ *(( ?>)+) *(.*)?")

CALLOUT_MAPPING = {
    "abstract": ["abstract", "summary", "tldr"],
    "bug": ["bug"],
    "example": ["example"],
    "danger": ["danger", "error"],
    "failure": ["fail", "failure", "missing"],
    "info": ["info"],
    "note": ["note"],
    "question": ["faq", "help", "question"],
    "quote": ["cite", "quote"],
    "tip": ["hint", "important", "tip"],
    "todo": ["todo"],
    "success": ["check", "done", "success"],
    "warning": ["attention", "caution", "warning"],
    "settings": ["settings"],
    "yaml": ["yaml"],
}
ADMONITION_MAPPING = {
    callout: admonition for admonition, callouts in CALLOUT_MAPPING.items() for callout in callouts
}
INDENTATION_MAPPING = {
    "tabs": "\t",
    "spaces": "    ",
}
FOLDABILITY_MAPPING = {
    "": "!!!",
    "-": "???",
    "+": "???+",
}
ALIGN_MAPPING = {"left": " inline", "right": " inline end"}


class CalloutToAdmonition:
    def __init__(self, callouts_config: _ObsidianCalloutsConfig):
        self._callouts_config: _ObsidianCalloutsConfig = callouts_config
        self._current_file_path: Optional[str] = None

    @staticmethod
    def _callout_indentation(
        match: re.Match, match_group: int, text_indentation: str = "spaces"
    ) -> str:
        initial_indentation = "".join(" " for _ in range(match.start(match_group)))
        indentation = "".join(
            INDENTATION_MAPPING.get(text_indentation, "spaces")
            for _ in range(match.group(match_group).count(">") - 1)
        )
        return f"{initial_indentation}{indentation}"

    def _callout_block_follow(self, match: re.Match, text_indentation: str = "spaces") -> str:
        indentation = self._callout_indentation(
            match=match, match_group=1, text_indentation=text_indentation
        )
        return f"    {indentation} {match.group(3)}\n"

    def _callout_block(self, match: re.Match, text_indentation: str = "spaces") -> str:
        # TODO: add possibility to inject customized callout/admonition from CSS/config

        indentation = self._callout_indentation(
            match=match, match_group=1, text_indentation=text_indentation
        )

        callout_type = match.group(3).lower()
        if "|" in callout_type:
            callout_type, align = callout_type.split("|")
            align = ALIGN_MAPPING.get(align)
            if align is None:
                log.warning(
                    f'Wrong alignment type "{align}" in file: {self._current_file_path} '
                    f"(fallback to no alignment, possible values {ALIGN_MAPPING.keys()})"
                )
                align = ""
        else:
            align = ""

        admonition_type = ADMONITION_MAPPING.get(callout_type)
        if admonition_type is None:
            log.warning(
                f'There is no callout mapping for "{match.group(3)}" '
                f'in file: {self._current_file_path} (fallback to "note")'
            )
            admonition_type = "note"

        foldable = FOLDABILITY_MAPPING.get(match.group(4))
        if foldable is None:
            log.warning(
                f'Wrong definition of foldable "{match.group(4)}" '
                f"in file: {self._current_file_path} (fallback to non foldable)"
            )
            foldable = "!!!"

        title = match.group(5)
        if title and title not in ['""', "''"]:
            title = f'"{title}"'

        return f"{indentation}{foldable} {admonition_type}{align} {title}\n"

    def convert_callouts(self, markdown: str, file_path: str) -> str:
        self._current_file_path = file_path
        in_callout_block: bool = False
        markdown_lines = []
        for line in markdown.split("\n"):
            callout_match = re.match(CALLOUT_BLOCK, line)
            callout_follow_match = re.match(CALLOUT_BLOCK_FOLLOW, line)
            if callout_match:
                in_callout_block = True
                markdown_lines.append(
                    self._callout_block(
                        match=callout_match, text_indentation=self._callouts_config.indentation
                    )
                )
            elif in_callout_block and callout_follow_match:
                markdown_lines.append(
                    self._callout_block_follow(
                        match=callout_follow_match,
                        text_indentation=self._callouts_config.indentation,
                    )
                )
            else:
                in_callout_block = False
                markdown_lines.append(line)

        return "\n".join(line for line in markdown_lines)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mkdocs.config import config_options as option
from mkdocs.config.base import Config


class _SocialOpenGraphConfig(Config):
    enabled = option.Type(bool, default=True)
    locale = option.Type(str, default="en_US")


class _SocialTwitterConfig(Config):
    enabled = option.Type(bool, default=True)
    website = option.Type(str, default="")
    author = option.Type(str, default="")


class _SocialMetaKeysConfig(Config):
    title_key = option.Type(str, default="title")
    description_key = option.Type(str, default="description")
    image_key = option.Type(str, default="image")


class SocialConfig(Config):
    twitter: _SocialTwitterConfig = option.SubConfig(_SocialTwitterConfig)  # type: ignore
    og: _SocialOpenGraphConfig = option.SubConfig(_SocialOpenGraphConfig)  # type: ignore
    meta_keys: _SocialMetaKeysConfig = option.SubConfig(_SocialMetaKeysConfig)  # type: ignore



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import importlib.resources
import logging
from dataclasses import fields
from pathlib import Path
from typing import cast

import yaml

from mkdocs_publisher.blog import lang as lang_path
from mkdocs_publisher.blog.config import BlogPluginConfig
from mkdocs_publisher.blog.structures import Translation

log = logging.getLogger("mkdocs.plugins.publisher.blog.translate")


class Translate:
    def __init__(self, config: BlogPluginConfig):
        self._config: BlogPluginConfig = config
        self._translation: Translation = cast(Translation, None)

        self._read_lang()

    def _read_lang(self):
        lang_yaml_oath = Path(
            str(importlib.resources.files(lang_path).joinpath(f"{self._config.lang}.yaml"))
        )
        if not lang_yaml_oath.exists():
            log.warning(
                f"There is no translation for '{self._config.lang}' language, "
                f"so default language ('{BlogPluginConfig.lang.default}') will be used"
            )
            lang_yaml_oath = str(
                importlib.resources.files(lang_path).joinpath(
                    f"{BlogPluginConfig.lang.default}.yaml"
                )
            )
        with open(lang_yaml_oath) as lang_yaml:
            translation_yaml_data = yaml.safe_load(lang_yaml)
        translation_keys = [f.name for f in fields(Translation)]
        translation_data = {
            k: v for k, v in translation_yaml_data.items() if k in translation_keys
        }
        # Inject overrides from mkdocs.yml config
        for key in translation_keys:
            value = self._config.translation.get(key, None)
            if value is not None:
                translation_data[key] = value
        self._translation = Translation(**translation_data)

    @property
    def translation(self) -> Translation:
        return self._translation



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
import re
from typing import Optional

import jinja2

from mkdocs_publisher.obsidian.config import _ObsidianVegaConfig

log = logging.getLogger("mkdocs.plugins.publisher.obsidian.vega")


VEGA_BLOCK_START = re.compile(r"^( *)```(vega-lite|vega)")
VEGA_BLOCK_STOP = re.compile(r"^( *)```")
VEGA_CHART_TEMPLATE = """
<div id="vega-chart-{{ vega_chart_id }}"></div>
<script type="text/javascript">
  var vegaChart{{ vega_chart_id }} = {{ vega_chart }};
  vegaEmbed('#vega-chart-{{ vega_chart_id }}', vegaChart{{ vega_chart_id }})
  .then(result => console.log(result))
  .catch(console.warn);
</script>
"""


class VegaCharts:
    def __init__(self, vega_config: _ObsidianVegaConfig):
        self._vega_config: _ObsidianVegaConfig = vega_config
        self._vega_schema_mapping: dict = {
            "vega": self._vega_config.vega_schema,
            "vega-lite": self._vega_config.vega_lite_schema,
        }
        self._vega_chart_id: int = 0
        self._vega_found: bool = False

    @property
    def is_vega(self) -> bool:
        return self._vega_found

    def generate_charts(self, markdown: str) -> str:
        in_vega_block: bool = False
        vega_block_lines: list = []
        vega_schema: Optional[str] = None
        markdown_lines = []

        for line in markdown.split("\n"):
            vega_start_match = re.match(VEGA_BLOCK_START, line)
            if not in_vega_block and vega_start_match:
                in_vega_block = True
                vega_schema = self._vega_schema_mapping[vega_start_match.group(2)]
            elif in_vega_block:
                vega_stop_match = re.match(VEGA_BLOCK_STOP, line)
                if vega_stop_match:
                    # Create chart data as JSON
                    vega_chart_json = json.loads("\n".join(vega_block_lines))
                    if "$schema" not in vega_chart_json:
                        vega_chart_json["$schema"] = vega_schema

                    # Render chart
                    self._vega_chart_id += 1
                    vega_chart_context = {
                        "vega_chart_id": self._vega_chart_id,
                        "vega_chart": json.dumps(vega_chart_json),
                    }
                    vega_chart_template = jinja2.Environment(
                        loader=jinja2.BaseLoader()
                    ).from_string(VEGA_CHART_TEMPLATE)

                    vega_chart = vega_chart_template.render(vega_chart_context)

                    for vega_chart_line in vega_chart.split("\n"):
                        markdown_lines.append(vega_chart_line)

                    # Mark that vega is in that page
                    self._vega_found = True

                    # Restore default values
                    in_vega_block = False
                    vega_block_lines = []
                    vega_schema = None
                else:
                    vega_block_lines.append(line)
            else:
                markdown_lines.append(line)

        return "\n".join(line for line in markdown_lines)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
from pathlib import Path

import click
import yaml
from mkdocs.utils import meta as meta_parser

from mkdocs_publisher._shared import mkdocs_utils
from mkdocs_publisher.obsidian.plugin import ObsidianPlugin

log = logging.getLogger("mkdocs.plugins.publisher.cli.md_tools")


@click.group
def app():
    """Markdown tools."""
    pass


@app.command()
def key_rename():
    mkdocs_config = mkdocs_utils.get_mkdocs_config()
    obsidian_plugin = ObsidianPlugin()
    obsidian_plugin.load_config(options={"plugins": mkdocs_config.plugins})

    obsidian_config_dir = Path(mkdocs_config.docs_dir) / obsidian_plugin.config.obsidian_dir
    linter_config_file = Path(obsidian_config_dir) / "plugins/obsidian-linter/data.json"

    key_order_list = []
    if linter_config_file.is_file():
        with linter_config_file.open("r") as linter_json:
            linter_config = json.load(linter_json)["ruleConfigs"]
            key_order_list = str(
                linter_config["yaml-key-sort"]["yaml-key-priority-sort-order"]
            ).split("\n")

    # old_key_name = click.prompt("Enter old key name you want to replace", type=str)
    old_key_name = "visibility"
    # new_key_name = click.prompt("Enter new key name", type=str)
    new_key_name = "publish"

    for file_path in Path(mkdocs_config.docs_dir).glob("**/*.md"):
        output = "---\n"
        with file_path.open(encoding="utf-8-sig", errors="strict", mode="r+") as md_file:
            content, post_meta = meta_parser.get_data(md_file.read())
            for key in key_order_list:
                if key in post_meta and key in post_meta:
                    output = f"{output}{yaml.safe_dump({key: post_meta[key]}, indent=2)}"
                    post_meta.pop(key, None)
                elif key in [old_key_name, new_key_name] and old_key_name in post_meta:
                    yaml_text_dump = yaml.safe_dump(
                        {new_key_name: post_meta[old_key_name]}, indent=2
                    )
                    output = f"{output}{yaml_text_dump}"
                    post_meta.pop(old_key_name, None)
            for key in post_meta:
                output = f"{output}{yaml.safe_dump({key: post_meta[key]}, indent=2)}"
            output = f"{output}---\n\n{content}"
            md_file.seek(0)
            md_file.write(output)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from dataclasses import dataclass
from hashlib import md5
from pathlib import Path
from typing import Optional

log = logging.getLogger("mkdocs.plugins.publisher._shared.links")


ANCHOR_RE_PART = r"((#(?P<anchor>([^|\][()'\"]+)))?)"
EXTRA_RE_PART = r"( *({(?P<extra>[\w+=. ]+)})?)"
IMAGE_RE_PART = r"((\|(?P<image>([0-9x]+)))?)"
LINK_RE_PART = r"(?P<link>(?!(https?|ftp)://)[^|#()\r\n\t\f\v]+)"
URL_RE_PART = r"(?P<link>((https?|ftp)://)?[\w\-]{2,}\.[\w\-]{2,}(\.[\w\-]{2,})?([^\s\][)(]*))"
TEXT_RE_PART = r"(?P<text>[^\][|]+)"
LINK_TITLE_RE_PART = r"(( \"(?P<title>[ \S]+)\")?)"

HTTP_LINK_RE = re.compile(rf"\[{TEXT_RE_PART}]\({URL_RE_PART}\)")
WIKI_LINK_RE = re.compile(rf"(?<!!)\[\[{LINK_RE_PART}{ANCHOR_RE_PART}(\|{TEXT_RE_PART})?]]")
WIKI_EMBED_LINK_RE = re.compile(
    rf"!\[\[{LINK_RE_PART}{ANCHOR_RE_PART}{IMAGE_RE_PART}]]{EXTRA_RE_PART}"
)
MD_LINK_RE = re.compile(
    rf"(?<!!)\[{TEXT_RE_PART}]\({LINK_RE_PART}{ANCHOR_RE_PART}"
    rf"{LINK_TITLE_RE_PART}\){EXTRA_RE_PART}"
)
MD_EMBED_LINK_RE = re.compile(
    rf"!\[{TEXT_RE_PART}]\({LINK_RE_PART}{LINK_TITLE_RE_PART}\){EXTRA_RE_PART}"
)
RELATIVE_LINK_RE = re.compile(
    rf"\[{TEXT_RE_PART}?]\({LINK_RE_PART}{ANCHOR_RE_PART}{LINK_TITLE_RE_PART}\)"
)
ANCHOR_LINK_RE = re.compile(rf"(?<!!)\[{TEXT_RE_PART}]\({ANCHOR_RE_PART}{LINK_TITLE_RE_PART}\)")


class RelativePathFinder:
    def __init__(self, current_file_path: Path, docs_dir: Path, relative_path: Path):
        self._current_file_path: Path = current_file_path
        self._docs_dir: Path = docs_dir
        self._relative_path: Path = relative_path

    @property
    def current_file_path(self) -> Path:
        return self._current_file_path

    @property
    def relative_path(self) -> Path:
        return self._relative_path

    def get_full_file_path(self, file_path: Path) -> Optional[Path]:
        """Find full file path."""
        full_file_path = self._docs_dir / file_path
        log.debug(f"Looking for file: {str(full_file_path)}")
        if not full_file_path.is_file():
            # Build list of unique found files paths
            found_files_list: list[Path] = []
            for f in self._docs_dir.glob(f"**/{file_path}"):
                f = f.resolve(strict=True)
                if f not in found_files_list:  # pragma: no cover
                    found_files_list.append(f)
            # Check how many files were found (there should be only one)
            no_of_found_files = len(found_files_list)
            if no_of_found_files == 1:
                full_file_path = found_files_list[0]
                log.debug(f'File: "{file_path} found: "{full_file_path}"')
            elif no_of_found_files <= 0:
                log.error(f'File: "{file_path}" doesn\'t exists (from: "{self._docs_dir}")')
                full_file_path = None
            else:
                log.error(
                    f"Too much files found: "
                    f"{[str(f.relative_to(self._docs_dir)) for f in found_files_list]} "
                    f"for link in file: {str(full_file_path)}"
                )
                full_file_path = None
        return full_file_path

    def get_relative_file_path(self, file_path: Optional[Path]) -> Optional[str]:
        """Get file path relative to currently opened file."""
        if file_path is not None:
            current_file_parts = list(Path(self._current_file_path).parts)
            current_file_parts[0] = str(self._relative_path)
            found_file_parts = list(file_path.relative_to(self._docs_dir).parts)
            relative_file_parts = []
            relative_file_missing_pieces = found_file_parts[:]

            index = 0
            for index, part in enumerate(current_file_parts[:-1]):
                try:
                    if part == found_file_parts[index]:
                        del relative_file_missing_pieces[0]
                    else:
                        relative_file_parts.append("..")
                except IndexError:
                    relative_file_parts.append("..")
                    relative_file_parts.extend(relative_file_missing_pieces)
            if index < len(found_file_parts):
                relative_file_parts.extend(relative_file_missing_pieces)
            return str(Path(*relative_file_parts))
        else:
            return None


@dataclass
class LinkMatch:
    anchor: Optional[str]
    text: Optional[str]
    extra: Optional[str] = None
    link: Optional[str] = None
    title: Optional[str] = None
    is_wiki: bool = False

    def __repr__(self):
        anchor = f"#{self.anchor}" if self.anchor else ""
        extra: list = self.extra.strip().split(" ") if self.extra else []
        title = f' "{self.title}"' if self.title else ""

        if self.is_wiki:
            if self.text is None and self.anchor:
                self.text = f"{self.link} > {self.anchor}"
            elif self.text is None:
                self.text = self.link
            link = f"{self.link}.md" if self.link else ""
        else:
            link = self.link if self.link else ""

        link_extra = f'{{{" ".join(extra)}}}' if extra else ""
        final_link = f"[{self.text}]({link}{anchor}{title}){link_extra}"
        log.debug(final_link)
        return final_link

    @property
    def backlink_anchor(self):
        return f"#{md5(self.link.encode()).hexdigest()}"

    @property
    def as_backlink(self):
        anchor = f"#{self.anchor}" if self.anchor else ""
        extra: list = self.extra.strip().split(" ") if self.extra else []
        extra.append(self.backlink_anchor)
        title = f' "{self.title}"' if self.title else ""
        link = self.link if self.link else ""

        link_extra = f'{{{" ".join(extra)}}}' if extra else ""
        final_link = f"[{self.text}]({link}{anchor}{title}){link_extra}"
        log.debug(final_link)
        return final_link


@dataclass
class WikiEmbedLinkMatch:
    anchor: Optional[str]
    extra: Optional[str]
    image: Optional[str]
    link: str

    def __repr__(self):
        extra: list = self.extra.strip().split(" ") if self.extra else []

        if self.image:
            try:
                width, height = self.image.split("x")
            except ValueError:
                width = self.image
                height = None
            extra.append(f"width={width}")
            if height:
                extra.append(f"height={height}")

        if str(self.link).lower().endswith(".pdf"):
            extra.append("pdfjs")
            if self.anchor:
                extra.append(self.anchor)
            self.anchor = None

        link_extra = f'{{{" ".join(extra)}}}' if extra else ""
        final_link = f"![{self.link.split('/')[-1]}]({self.link}){link_extra}"
        log.debug(final_link)
        return final_link


@dataclass
class MdEmbedLinkMatch:
    extra: Optional[str]
    link: str
    text: str
    title: Optional[str]
    is_loading_lazy: bool = True

    def __repr__(self):
        extra: list = self.extra.strip().split(" ") if self.extra else []
        title = f' "{self.title}"' if self.title else ""

        if self.is_loading_lazy and "loading=lazy" not in extra:
            extra.append("loading=lazy")
        link_extra = f'{{{" ".join(extra)}}}' if extra else ""
        final_link = f"![{self.text}]({self.link}{title}){link_extra}"
        log.debug(final_link)
        return final_link


@dataclass
class RelativeLinkMatch:
    anchor: Optional[str]
    link: str
    text: str
    title: Optional[str]
    relative_path_finder: Optional[RelativePathFinder] = None

    def __repr__(self):
        anchor = f"#{self.anchor}" if self.anchor else ""
        title = f' "{self.title}"' if self.title else ""

        # The same page anchor link doesn't have file part
        if self.link.startswith("#"):
            final_link = f"[{self.text}]({self.link})"
            log.debug(final_link)
            return final_link

        # Link from blog sub-pages have to be recalculated for a new relative value
        if (
            str(self.relative_path_finder.current_file_path).startswith(
                str(self.relative_path_finder.relative_path)
            )
            or str(self.relative_path_finder.current_file_path).startswith("index-")
        ) and (
            # TODO: rethink RSS file filtering
            not self.link.endswith(".xml")  # RSS feed exclusion
        ):
            file_path = self.relative_path_finder.get_full_file_path(file_path=Path(self.link))
            if file_path is not None:
                link = str(self.relative_path_finder.get_relative_file_path(file_path=file_path))
            else:
                link = ""
        else:
            link = self.link
        final_link = f"[{self.text}]({link}{anchor}{title})"

        return final_link



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import contextlib
import logging
import re
from collections import OrderedDict
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Optional
from typing import cast

from mkdocs.config import Config
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.plugins import event_priority
from mkdocs.structure.files import Files
from mkdocs.structure.nav import Navigation
from mkdocs.structure.pages import Page

# noinspection PyProtectedMember
from mkdocs_publisher._shared import file_utils
from mkdocs_publisher._shared import links
from mkdocs_publisher._shared import resources
from mkdocs_publisher.blog import creators
from mkdocs_publisher.blog import modifiers
from mkdocs_publisher.blog import parsers
from mkdocs_publisher.blog.config import BlogPluginConfig
from mkdocs_publisher.blog.structures import BlogConfig
from mkdocs_publisher.obsidian.md_links import MarkdownLinks

log = logging.getLogger("mkdocs.plugins.publisher.blog.plugin")


class BlogPlugin(BasePlugin[BlogPluginConfig]):
    def __init__(self):
        self.blog_config = BlogConfig()  # Empty instance
        self._start_page: bool = False
        self._on_serve: bool = False

    def on_startup(self, *, command: Literal["build", "gh-deploy", "serve"], dirty: bool) -> None:
        if command == "serve":
            self._on_serve = True

    def on_config(self, config: MkDocsConfig) -> Config:
        # Initialization of all the values
        self.blog_config.parse_configs(mkdocs_config=config, plugin_config=self.config)

        # Modify nav section
        if config.nav is None:
            config.nav = [{str(self.blog_config.blog_dir): str(self.blog_config.blog_dir)}]

        # Detect if blog is a starting page
        if len(config.nav) > 0:
            first_value = list(config.nav[0].values())[0]
            if isinstance(first_value, str) and first_value == str(self.blog_config.blog_dir):
                self._start_page = True

        # New config navigation
        config_nav = OrderedDict()
        parsers.parse_markdown_files(
            blog_config=self.blog_config,
            config_nav=config_nav,
            on_serve=self._on_serve,
        )

        parsers.create_blog_post_teaser(
            blog_config=self.blog_config,
        )

        creators.create_blog_post_pages(
            start_page=self._start_page,
            blog_config=self.blog_config,
            config_nav=config_nav,
        )

        modifiers.blog_post_nav_sorter(
            blog_config=self.blog_config,
            config_nav=config_nav,
        )

        # Inject blog into navigation
        new_nav = []
        for nav_item in cast(list, config.nav):
            if str(self.blog_config.blog_dir) in nav_item.values():
                for k, v in config_nav.items():
                    new_nav.append({k: v})
            else:
                new_nav.append(nav_item)
        config.nav = new_nav

        return config

    def on_nav(self, nav: Navigation, config: MkDocsConfig, files: Files) -> Navigation:
        modifiers.blog_post_nav_remove(
            start_page=self._start_page, blog_config=self.blog_config, nav=nav
        )

        return nav

    def on_files(self, files: Files, config: MkDocsConfig) -> Files:
        creators.create_blog_files(blog_config=self.blog_config, files=files)

        resources.add_extra_css(stylesheet_file_name="blog.min.css", config=config, files=files)

        return files

    @event_priority(-100)  # Run after all other plugins
    def on_page_context(
        self, context: dict[str, Any], *, page: Page, config: MkDocsConfig, nav: Navigation
    ) -> Optional[dict[str, Any]]:
        if Path(page.file.src_path).parts[0] == self.config.blog_dir:
            page.meta[self.config.comments.key_name] = self.config.comments.enabled

        # Temporary created files cannot be edited
        if page.file.src_uri in self.blog_config.temp_files_list:
            page.edit_url = None

        modifiers.blog_post_nav_next_prev_change(
            start_page=self._start_page, blog_config=self.blog_config, page=page
        )
        return context

    @event_priority(-100)  # Run after all other plugins
    def on_page_markdown(
        self, markdown: str, *, page: Page, config: MkDocsConfig, files: Files
    ) -> Optional[str]:
        md_links = MarkdownLinks(mkdocs_config=config)
        markdown = md_links.normalize_relative_links(
            markdown=markdown, current_file_path=page.file.src_path
        )

        # Dirty hack for blog standalone mode index file
        if page.file.src_path == "index.md":

            def _blog_index_re(match: re.Match):
                blog_link = links.LinkMatch(**match.groupdict())

                relative_path_finder = links.RelativePathFinder(
                    current_file_path=Path(page.file.src_path),
                    docs_dir=Path(config.docs_dir),
                    relative_path=Path(config.docs_dir),
                )
                full_blog_link = relative_path_finder.get_full_file_path(
                    file_path=Path(str(blog_link.link))
                )
                blog_link.link = relative_path_finder.get_relative_file_path(
                    file_path=full_blog_link
                )
                return str(blog_link)

            markdown = re.sub(links.RELATIVE_LINK_RE, _blog_index_re, markdown)
        return markdown

    @event_priority(-100)  # Run after all other plugins
    def on_build_error(self, error: Exception) -> None:
        with contextlib.suppress(AttributeError):
            file_utils.remove_dir(directory=self.blog_config.temp_dir)

    @event_priority(-100)  # Run after all other plugins
    def on_shutdown(self) -> None:
        with contextlib.suppress(AttributeError):
            file_utils.remove_dir(directory=self.blog_config.temp_dir)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from pathlib import Path
from typing import Optional

from mkdocs.config.defaults import MkDocsConfig

# noinspection PyProtectedMember
from mkdocs_publisher._shared import mkdocs_utils
from mkdocs_publisher.blog.config import BlogPluginConfig
from mkdocs_publisher.meta.config import MetaPluginConfig


@dataclass
class BlogPost:
    """Single blog post data container"""

    title: str
    date: datetime
    path: Optional[str]
    content: Optional[str]
    tags: Optional[list[str]]
    categories: Optional[list[str]]
    slug: Optional[str] = None
    teaser: str = ""
    is_teaser: bool = False

    @property
    def as_dict(self) -> dict:
        return asdict(self)


@dataclass
class Translation:
    teaser_link_text: str
    blog_page_title: str
    blog_navigation_name: str
    recent_blog_posts_navigation_name: str
    archive_page_title: str
    archive_navigation_name: str
    categories_page_title: str
    categories_navigation_name: str
    tags_page_title: str
    tags_navigation_name: str
    newer_posts: str
    older_posts: str

    @property
    def as_dict(self) -> dict:
        return asdict(self)


@dataclass
class BlogConfig:
    mkdocs_config: MkDocsConfig = field(init=False)
    plugin_config: BlogPluginConfig = field(init=False)
    meta_config: Optional[MetaPluginConfig] = field(init=False, default=None)
    translation: Translation = field(init=False)
    temp_dir: Path = field(init=False)
    docs_dir: Path = field(init=False)
    blog_dir: Path = field(init=False)
    site_dir: Path = field(init=False)
    blog_posts: dict[datetime, BlogPost] = field(init=False, default_factory=lambda: dict())
    temp_files: dict[str, Path] = field(init=False, default_factory=lambda: dict())

    @property
    def temp_files_list(self) -> list[str]:
        temp_files = []
        for path in self.temp_files.values():
            temp_files.append(str(path.relative_to(self.temp_dir)))
        return temp_files

    def parse_configs(self, mkdocs_config: MkDocsConfig, plugin_config: BlogPluginConfig):
        from mkdocs_publisher.blog.translate import Translate

        self.mkdocs_config = mkdocs_config
        self.plugin_config = plugin_config
        self.meta_config: Optional[MetaPluginConfig] = (
            mkdocs_utils.get_plugin_config(mkdocs_config=mkdocs_config, plugin_name="pub-meta")
            or None
        )  # type: ignore
        self.temp_dir = Path(plugin_config.temp_dir)
        self.docs_dir = Path(mkdocs_config.docs_dir)
        self.blog_dir = Path(plugin_config.blog_dir)
        self.site_dir = Path(mkdocs_config.site_dir)
        self.translation = Translate(config=plugin_config).translation

        self.temp_dir.mkdir(exist_ok=True)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import platform
import sys
from io import BytesIO
from pathlib import Path
from typing import cast
from zipfile import ZIP_DEFLATED
from zipfile import ZipFile

from mkdocs.plugins import BasePlugin
from mkdocs.plugins import event_priority

# noinspection PyProtectedMember
from mkdocs_publisher._shared import file_utils
from mkdocs_publisher._shared import mkdocs_utils
from mkdocs_publisher.debugger import loggers
from mkdocs_publisher.debugger.config import DebuggerConfig

log = logging.getLogger("mkdocs.plugins.publisher.debug.plugin")

LOG_FILENAME_SUFFIX = "_mkdocs_build.log"
ZIP_FILENAME_SUFFIX = "_mkdocs_debug.zip"
FILES_TO_ZIP_LIST = [
    "mkdocs.yml",
    "requirements.txt",
    "pyproject.toml",
    "poetry.lock",
    ".gitignore",
]
PIP_FREEZE_FILENAME = "requirements_freeze.txt"


# noinspection PyProtectedMember
class DebuggerPlugin(BasePlugin[DebuggerConfig]):
    def __init__(self):
        self._mkdocs_log_stream_handler: logging.Handler = logging.getLogger("mkdocs").handlers[0]

        self._mkdocs_log_file_handler: loggers.DatedFileHandler = loggers.DatedFileHandler(
            filename=f"%Y%m%d_%H%M%S{LOG_FILENAME_SUFFIX}"
        )

        self._mkdocs_log_file: str = str(Path(self._mkdocs_log_file_handler.baseFilename).name)
        self._mkdocs_log_date: str = self._mkdocs_log_file.replace(LOG_FILENAME_SUFFIX, "")

        self.load_config(
            options=cast(
                dict,
                mkdocs_utils.get_plugin_config(
                    mkdocs_config=mkdocs_utils.get_mkdocs_config(), plugin_name="pub-debugger"
                ),
            ),
            config_file_path=mkdocs_utils.get_mkdocs_config().config_file_path,
        )

        if self.config.console_log.enabled:
            self._mkdocs_log_stream_handler.setFormatter(
                loggers.ProjectPathStreamFormatter(console_config=self.config.console_log)
            )
            self._mkdocs_log_stream_handler.addFilter(
                loggers.ProjectPathConsoleFilter(console_config=self.config.console_log)
            )
            # noinspection PyUnresolvedReferences
            self._mkdocs_log_stream_handler.level = logging._nameToLevel[
                self.config.console_log.log_level
            ]

            logging.getLogger("root").handlers = [self._mkdocs_log_stream_handler]

        if self.config.file_log.enabled:
            self._mkdocs_log_file_handler.setFormatter(
                loggers.ProjectPathFileFormatter(fmt=self.config.file_log.log_format)
            )
            self._mkdocs_log_file_handler.addFilter(
                loggers.ProjectPathFileFilter(file_config=self.config.file_log)
            )
            # noinspection PyUnresolvedReferences
            self._mkdocs_log_file_handler.level = logging._nameToLevel[
                self.config.file_log.log_level
            ]

            logging.getLogger("mkdocs").handlers.append(self._mkdocs_log_file_handler)

            if self.config.file_log.remove_old_files:
                for log_file in Path(".").rglob(f"*{LOG_FILENAME_SUFFIX}"):
                    if str(log_file) != self._mkdocs_log_file:
                        log_file.unlink(missing_ok=True)

    @event_priority(100)  # Run after  all other plugins
    def on_shutdown(self) -> None:
        if self.config.file_log.enabled:
            log.info(f"Platform: {platform.platform()}")
            log.info(
                f"Python version: {platform.python_version()} "
                f"(using virtual environment: {str(sys.prefix != sys.base_prefix).lower()})"
            )
            log.info(f"Build log file: {self._mkdocs_log_file_handler.baseFilename}")

        if self.config.zip_log.enabled:
            zip_file_name = f"{self._mkdocs_log_date}{ZIP_FILENAME_SUFFIX}"

            if self.config.zip_log.remove_old_files:
                for log_file in Path(".").rglob(f"*{ZIP_FILENAME_SUFFIX}"):
                    if str(log_file) != zip_file_name:
                        log_file.unlink(missing_ok=True)

            archive = BytesIO()
            with ZipFile(archive, "a", ZIP_DEFLATED, False) as archive_file:
                # Zip files from list
                for file_to_zip in FILES_TO_ZIP_LIST:
                    if Path(file_to_zip).exists():
                        log.debug(f"File: {file_to_zip} added to archive")
                        archive_file.write(filename=file_to_zip, arcname=file_to_zip)
                    else:
                        log.debug(f"File: {file_to_zip} doesn't exists and not added to archive")

                # Write pip freeze as file
                if self.config.zip_log.add_pip_freeze:
                    pip_run = file_utils.run_subprocess(["pip", "freeze", "--local"])
                    archive_file.writestr(
                        zinfo_or_arcname=PIP_FREEZE_FILENAME,
                        data=pip_run.stdout.decode("utf-8"),
                    )

                # Write build log file
                archive_file.write(filename=self._mkdocs_log_file, arcname=self._mkdocs_log_file)

            with open(zip_file_name, "wb") as zip_file:
                zip_file.write(archive.getvalue())
                log.info(f"Debugger ZIP file: {zip_file_name}")



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

from bs4 import BeautifulSoup

log = logging.getLogger("mkdocs.plugins.publisher._shared.html_modifiers")


class HTMLModifier:
    """HTML modifications"""

    def __init__(self, markup: str):
        self._soup = BeautifulSoup(markup=markup, features="html.parser")

    def __repr__(self):
        return self._soup.prettify()

    def add_head_script(self, src: str):
        """Add script to the head section"""
        self._soup.head.append(  # type: ignore
            self._soup.new_tag(name="script", attrs={"src": src})
        )

    def add_meta_property(self, name: str, value: str):
        """Add property to the head section"""
        self._soup.head.append(  # type: ignore
            self._soup.new_tag(name="meta", attrs={"property": name, "content": value})
        )

    def remove_meta_properties(self, properties: list[str]):
        """Remove meta tags with given properties from head section"""
        for prop in properties:
            head_property = self._soup.head.find(name="meta", attr={"property": prop})  # type: ignore
            if head_property is not None:
                head_property.extract()

    def fix_img_links(self):
        images = self._soup.findAll(name="img")
        for img in images:
            img_src = str(img.attrs["src"])
            if img_src.endswith("/"):
                img["src"] = img_src[:-1]



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import importlib.resources
import logging
from pathlib import Path

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.files import File
from mkdocs.structure.files import Files

from mkdocs_publisher._extra import assets

# noinspection PyProtectedMember
from mkdocs_publisher._extra.assets import stylesheets
from mkdocs_publisher._extra.assets import templates

log = logging.getLogger("mkdocs.plugins.publisher._shared.resources")


def _add_extra_file(
    resource_file_path: Path,
    site_dir: str,
    use_directory_urls: bool,
    config_extra_files: list,
    files: Files,
):
    assets_path = Path(str(importlib.resources.files(assets)))
    extra_file_path = str(resource_file_path.relative_to(assets_path.parent))
    if resource_file_path.exists():
        files.append(
            File(
                path=extra_file_path,
                src_dir=str(assets_path.parent),
                dest_dir=str(site_dir),
                use_directory_urls=use_directory_urls,
            )
        )
        config_extra_files.append(extra_file_path)
        log.debug(f"Extra file added: {extra_file_path}")
    else:
        log.error(f"Extra file doesn't exists: {extra_file_path}")


def add_extra_css(
    stylesheet_file_name: str, config: MkDocsConfig, files: Files, add_map: bool = True
):
    """Add CSS file from mkdocs_publisher._extra to mkdocs.yml config file"""

    css_file_path = Path(
        str(importlib.resources.files(stylesheets).joinpath(stylesheet_file_name))
    )
    _add_extra_file(
        resource_file_path=css_file_path,
        site_dir=config.site_dir,
        use_directory_urls=config.use_directory_urls,
        config_extra_files=config.extra_css,
        files=files,
    )

    if add_map:
        css_map_file_path = Path(
            str(importlib.resources.files(stylesheets).joinpath(f"{stylesheet_file_name}.map"))
        )
        _add_extra_file(
            resource_file_path=css_map_file_path,
            site_dir=config.site_dir,
            use_directory_urls=config.use_directory_urls,
            config_extra_files=config.extra_css,
            files=files,
        )


def read_template_file(template_file_name: str) -> str:
    """Read and return content of template file"""

    resource_file_path = importlib.resources.files(templates).joinpath(template_file_name)
    with importlib.resources.as_file(resource_file_path) as template_file:
        return template_file.read_text(encoding="utf-8")



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from multiprocessing import cpu_count
from pathlib import Path
from typing import Literal
from typing import cast

import yaml
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.plugins import event_priority

from mkdocs_publisher.minifier import minifiers
from mkdocs_publisher.minifier.base import CachedFile
from mkdocs_publisher.minifier.config import MinifierConfig

log = logging.getLogger("mkdocs.plugins.publisher.minifier.plugin")


class MinifierPlugin(BasePlugin[MinifierConfig]):
    def __init__(self):
        self._on_serve: bool = False

    def on_startup(self, *, command: Literal["build", "gh-deploy", "serve"], dirty: bool) -> None:
        if command == "serve":
            self._on_serve = True

    @event_priority(-100)  # Run after all other plugins
    def on_post_build(self, *, config: MkDocsConfig) -> None:
        Path(self.config.cache_dir).mkdir(exist_ok=True)

        # TODO: Add path to tools checker
        cached_files: dict[str, CachedFile] = {}

        if self.config.threads == 0:
            self.config.threads = int(cpu_count())
        if self.config.threads < 1:
            log.info('Number of "threads" cannot be smaller than 1 (changing to minimal 1)')
            self.config.threads = 1
        log.info(f"Threads used for minifiers: {self.config.threads}")

        cached_files_list: Path = Path(self.config.cache_dir) / self.config.cache_file
        if cached_files_list.exists():
            try:
                with open(cached_files_list) as yaml_file:
                    cached_files = yaml.safe_load(yaml_file)
                    for file_path, cached_file in cached_files.items():
                        cached_files[file_path] = CachedFile(**cast(dict, cached_file))
            except (yaml.YAMLError, AttributeError) as e:
                log.warning(f"File '{cached_files_list}' corrupted. Rebuilding cache.")
                log.debug(e)

        if self.config.png.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.png.enabled_on_serve)
        ):
            minifiers.PngMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.jpeg.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.jpeg.enabled_on_serve)
        ):
            minifiers.JpegMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.svg.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.svg.enabled_on_serve)
        ):
            minifiers.SvgMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.html.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.html.enabled_on_serve)
        ):
            minifiers.HtmlMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.css.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.css.enabled_on_serve)
        ):
            minifiers.CssMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.js.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.js.enabled_on_serve)
        ):
            minifiers.JsMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        cached_files_for_yaml = {}
        for file_name, cached_file in cached_files.items():
            # TODO: Cleanup yaml from unused/deleted files and delete those that are unused
            cached_file = {k: str(v) for k, v in cached_file.as_dict.items()}
            cached_files_for_yaml[file_name] = cached_file

        with open(cached_files_list, "w") as yaml_file:
            yaml.safe_dump(cached_files_for_yaml, yaml_file)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from pathlib import Path
from typing import Optional

# noinspection PyProtectedMember
from mkdocs_publisher._shared import file_utils
from mkdocs_publisher.minifier import config as minifier_config
from mkdocs_publisher.minifier.base import BaseMinifier
from mkdocs_publisher.minifier.base import CachedFile

log = logging.getLogger("mkdocs.plugins.publisher.minifier.minifiers")


class PngMinifier(BaseMinifier):
    def __call__(self):
        self._minify_options: minifier_config._MinifierPngConfig = self._plugin_config.png
        super().__call__()

    def minifier(self, cached_file: CachedFile) -> Optional[CachedFile]:
        try:
            input_file = self._mkdocs_config.site_dir / cached_file.original_file_path
            output_file = self._plugin_config.cache_dir / cached_file.cached_file_name

            if self._minify_options.pngquant_enabled:
                pngquant_cmd = [
                    self._minify_options.pngquant_path,
                    "--force",
                    "--strip" if self._minify_options.strip else None,
                    "--quality" if int(self._minify_options.pngquant_quality) > 0 else None,
                    self._minify_options.pngquant_quality
                    if int(self._minify_options.pngquant_quality) > 0
                    else None,
                    "--speed",
                    self._minify_options.pngquant_speed,
                    "--output",
                    str(output_file),
                    str(input_file),
                ]
                if file_utils.run_subprocess(cmd=pngquant_cmd).returncode != 0:
                    output_file.unlink(missing_ok=True)
                    return None

            if self._minify_options.oxipng_enabled:
                oxipng_cmd = [
                    self._minify_options.oxipng_path,
                    "--quiet",
                    "--force",
                    "--strip" if self._minify_options.strip else None,
                    "all" if self._minify_options.strip else None,
                    "--zopfli" if self._minify_options.oxipng_max_compression else None,
                    "--out" if not self._minify_options.pngquant_enabled else None,
                    str(output_file) if not self._minify_options.pngquant_enabled else None,
                    str(input_file)
                    if not self._minify_options.pngquant_enabled
                    else str(output_file),
                ]
                if file_utils.run_subprocess(cmd=oxipng_cmd).returncode != 0:
                    output_file.unlink(missing_ok=True)
                    return None

                return cached_file

        except FileNotFoundError as e:
            log.warning(e)

        return None


class JpegMinifier(BaseMinifier):
    def __call__(self):
        self._minify_options: minifier_config._MinifierJpegConfig = self._plugin_config.jpeg
        super().__call__()

    def minifier(self, cached_file: CachedFile) -> Optional[CachedFile]:
        try:
            input_file = self._mkdocs_config.site_dir / cached_file.original_file_path
            output_file = self._plugin_config.cache_dir / cached_file.cached_file_name

            djpg_cmd = [
                self._minify_options.djpeg_path,
                "-targa",
                "-outfile",
                str(output_file.with_suffix(".tga")),
                str(input_file),
            ]
            if file_utils.run_subprocess(cmd=djpg_cmd).returncode != 0:
                output_file.with_suffix(".tga").unlink(missing_ok=True)
                return None

            cjpg_cmd = [
                self._minify_options.cjpeg_path,
                "-targa",
                "-smooth" if int(self._minify_options.smooth) > 0 else None,
                self._minify_options.smooth if int(self._minify_options.smooth) > 0 else None,
                "-quality" if int(self._minify_options.quality) > 0 else None,
                self._minify_options.quality if int(self._minify_options.quality) > 0 else None,
                "-outfile",
                str(output_file.with_suffix(".jpg_")),
                str(output_file.with_suffix(".tga")),
            ]
            if file_utils.run_subprocess(cmd=cjpg_cmd).returncode != 0:
                output_file.with_suffix(".jpg_").unlink(missing_ok=True)
                return None
            output_file.with_suffix(".tga").unlink(missing_ok=True)

            jpegtran_cmd = [
                self._minify_options.jpegtran_path,
                "-copy",
                self._minify_options.copy_meta,
                "-progressive" if self._minify_options.progressive else None,
                "-optimise" if self._minify_options.optimise else None,
                "-outfile",
                str(output_file),
                str(output_file.with_suffix(".jpg_")),
            ]
            if file_utils.run_subprocess(cmd=jpegtran_cmd).returncode != 0:
                output_file.unlink(missing_ok=True)
                return None
            output_file.with_suffix(".jpg_").unlink(missing_ok=True)

            return cached_file

        except FileNotFoundError as e:
            log.warning(e)

        return None


class SvgMinifier(BaseMinifier):
    def __call__(self):
        self._minify_options: minifier_config._MinifierSvgConfig = self._plugin_config.svg
        super().__call__()

    def minifier(self, cached_file: CachedFile) -> Optional[CachedFile]:
        try:
            input_file = self._mkdocs_config.site_dir / cached_file.original_file_path
            output_file = self._plugin_config.cache_dir / cached_file.cached_file_name

            svgo_cmd = [
                self._minify_options.svgo_path,
                "--multipass" if self._minify_options.multipass else None,
                "--quiet",
                "--output",
                str(output_file),
                "--input",
                str(input_file),
            ]
            if file_utils.run_subprocess(cmd=svgo_cmd).returncode != 0:
                output_file.unlink(missing_ok=True)
                return None

            return cached_file

        except FileNotFoundError as e:
            log.warning(e)

        return None


class HtmlMinifier(BaseMinifier):
    def __call__(self):
        self._minify_options: minifier_config._MinifierHtmlConfig = self._plugin_config.html
        super().__call__()

    def minifier(self, cached_file: CachedFile) -> Optional[CachedFile]:
        try:
            input_file = self._mkdocs_config.site_dir / cached_file.original_file_path
            output_file = self._plugin_config.cache_dir / cached_file.cached_file_name

            html_minifier_cmd = [
                self._minify_options.html_minifier_path,
                "--case-sensitive" if self._minify_options.case_sensitive else None,
                "--minify-css" if self._minify_options.minify_css else None,
                "--minify-js" if self._minify_options.minify_js else None,
                "--remove-comments" if self._minify_options.remove_comments else None,
                "--remove-tag-whitespace" if self._minify_options.remove_tag_whitespace else None,
                "--collapse-whitespace" if self._minify_options.collapse_whitespace else None,
                "--conservative-collapse" if self._minify_options.conservative_collapse else None,
                "--collapse-boolean-attributes"
                if self._minify_options.collapse_boolean_attributes
                else None,
                "--preserve-line-breaks" if self._minify_options.preserve_line_breaks else None,
                "--max-line-length" if int(self._minify_options.max_line_length) > 0 else None,
                self._minify_options.max_line_length
                if int(self._minify_options.max_line_length) > 0
                else None,
                "--sort-attributes" if self._minify_options.sort_attributes else None,
                "--sort-class-name" if self._minify_options.sort_class_name else None,
                "--output",
                str(output_file),
                str(input_file),
            ]
            if file_utils.run_subprocess(cmd=html_minifier_cmd).returncode != 0:
                output_file.unlink(missing_ok=True)
                return None

            return cached_file

        except FileNotFoundError as e:
            log.warning(e)

        return None


class CssMinifier(BaseMinifier):
    def __call__(self):
        self._minify_options: minifier_config._MinifierCssConfig = self._plugin_config.css
        super().__call__()

    def minifier(self, cached_file: CachedFile) -> Optional[CachedFile]:
        try:
            input_file = self._mkdocs_config.site_dir / cached_file.original_file_path
            output_file = self._plugin_config.cache_dir / cached_file.cached_file_name

            if (
                ".min" in Path(str(input_file).lower()).suffixes
                and self._minify_options.skip_minified
            ):
                return None
            css_minifier_cmd = [
                self._minify_options.postcss_path,
                "--no-map",
                "--use",
                "cssnano",
                "postcss-svgo",
                "--output",
                str(output_file),
                str(input_file),
            ]
            if file_utils.run_subprocess(cmd=css_minifier_cmd).returncode != 0:
                output_file.unlink(missing_ok=True)
                return None

            return cached_file

        except FileNotFoundError as e:
            log.warning(e)

        return None


class JsMinifier(BaseMinifier):
    def __call__(self):
        self._minify_options: minifier_config._MinifierJsConfig = self._plugin_config.js
        super().__call__()

    def minifier(self, cached_file: CachedFile) -> Optional[CachedFile]:
        try:
            input_file = self._mkdocs_config.site_dir / cached_file.original_file_path
            output_file = self._plugin_config.cache_dir / cached_file.cached_file_name

            if (
                ".min" in Path(str(input_file).lower()).suffixes
                and self._minify_options.skip_minified
            ):
                return None
            js_minifier_cmd = [
                self._minify_options.uglifyjs_path,
                "--compress",
                "--webkit",
                "--output",
                str(output_file),
                str(input_file),
            ]
            if file_utils.run_subprocess(cmd=js_minifier_cmd).returncode != 0:
                output_file.unlink(missing_ok=True)
                return None

            return cached_file

        except FileNotFoundError as e:
            log.warning(e)

        return None



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import cast

import jinja2
import yaml
from mkdocs.structure.files import File
from mkdocs.structure.files import Files

# noinspection PyProtectedMember
from mkdocs_publisher._shared import resources

# noinspection PyProtectedMember
from mkdocs_publisher._shared.urls import slugify
from mkdocs_publisher.blog.structures import BlogConfig

log = logging.getLogger("mkdocs.plugins.publisher.blog.creators")


def create_blog_files(
    blog_config: BlogConfig,
    files: Files,
):
    for temp_file in blog_config.temp_files.values():
        try:
            file = File(
                path=str(Path(temp_file).relative_to(blog_config.temp_dir)),
                src_dir=str(blog_config.temp_dir),
                dest_dir=str(blog_config.site_dir),
                use_directory_urls=blog_config.mkdocs_config.use_directory_urls,
            )
            files.append(file)
        except ValueError:
            pass


def create_blog_post_pages(
    start_page: bool,
    blog_config: BlogConfig,
    config_nav: OrderedDict,
) -> None:
    """Create blog posts index files."""

    log.info("Creating blog posts index files")
    posts_chunks: dict[str, list] = {}
    archive_chunks: dict[str, list] = {}
    categories_chunks: dict[str, list] = {}
    tags_chunks: dict[str, list] = {}

    # Build post index pages
    for index, date in enumerate(sorted(blog_config.blog_posts, reverse=True)):
        index = (
            "index"
            if start_page and index < blog_config.plugin_config.posts_per_page
            else f"index-{str(index//blog_config.plugin_config.posts_per_page)}"
        )
        if index not in posts_chunks:
            posts_chunks[index] = []
        posts_chunks[index].append(blog_config.blog_posts[date])

    # Build archive, category and tag pages
    for date in sorted(blog_config.blog_posts, reverse=True):
        date: datetime
        year = str(date.year)
        if year not in archive_chunks:
            archive_chunks[year] = []
        archive_chunks[year].append(blog_config.blog_posts[date].as_dict)

        categories = cast(str, blog_config.blog_posts[date].categories)
        for category in categories:
            if category not in categories_chunks:
                categories_chunks[category] = []
            categories_chunks[category].append(blog_config.blog_posts[date].as_dict)

        tags = cast(list, blog_config.blog_posts[date].tags)
        for tag in tags:
            if tag not in tags_chunks:
                tag = tag.replace("/", "-")
                tags_chunks[tag] = []
            tags_chunks[tag].append(blog_config.blog_posts[date].as_dict)

    # Reorder categories alphabetically
    categories_chunks = {cat: categories_chunks[cat] for cat in sorted(categories_chunks)}

    # Reorder tags alphabetically
    tags_chunks = {tag: tags_chunks[tag] for tag in sorted(tags_chunks)}
    config_nav[blog_config.translation.blog_navigation_name] = []

    for key, single_posts_chunk in posts_chunks.items():
        file_name = f"{key}.md"
        file_path = blog_config.temp_dir / file_name

        _render_and_write_page(
            single_posts_chunk=single_posts_chunk,
            file_path=file_path,
            blog_config=blog_config,
            page_title=blog_config.translation.blog_page_title,
        )

        blog_config.temp_files[f"{blog_config.translation.blog_navigation_name}/{key}"] = file_path
        log.debug(f"Creating blog post chunk file: {file_path}")

        config_nav[blog_config.translation.blog_navigation_name].append({key: file_name})

    config_nav[blog_config.translation.blog_navigation_name].append(
        {
            blog_config.translation.archive_navigation_name: _create_pages(
                blog_config=blog_config,
                posts_chunks=archive_chunks,
                sub_dir=Path(blog_config.plugin_config.archive_subdir),
                page_title=blog_config.translation.archive_page_title,
            )
        }
    )

    config_nav[blog_config.translation.blog_navigation_name].append(
        {
            blog_config.translation.categories_navigation_name: _create_pages(
                blog_config=blog_config,
                posts_chunks=categories_chunks,
                sub_dir=Path(blog_config.plugin_config.categories_subdir),
                page_title=blog_config.translation.categories_page_title,
            )
        }
    )

    config_nav[blog_config.translation.blog_navigation_name].append(
        {
            blog_config.translation.tags_navigation_name: _create_pages(
                blog_config=blog_config,
                posts_chunks=tags_chunks,
                sub_dir=Path(blog_config.plugin_config.tags_subdir),
                page_title=blog_config.translation.tags_page_title,
            )
        }
    )


def _create_pages(
    blog_config: BlogConfig,
    posts_chunks: dict[str, list],
    sub_dir: Path,
    page_title: str,
) -> list[dict[str, str]]:
    config_nav = []

    blog_temp_dir = blog_config.temp_dir / blog_config.plugin_config.blog_dir
    blog_temp_dir.mkdir(exist_ok=True)

    pages_dir = blog_temp_dir / sub_dir
    pages_dir.mkdir(exist_ok=True)

    for key, single_posts_chunk in posts_chunks.items():
        file_name = f"{key}.md"
        file_path = pages_dir / file_name
        _render_and_write_page(
            single_posts_chunk=single_posts_chunk,
            file_path=file_path,
            blog_config=blog_config,
            page_title=f"{page_title} - {key}",
        )

        blog_config.temp_files[f"{sub_dir}/{key}"] = file_path
        log.debug(f"Creating blog post chunk file: {file_path}")

        config_nav.append({key: f"{blog_config.plugin_config.blog_dir}/{sub_dir}/{file_name}"})
    return config_nav


def _render_and_write_page(
    single_posts_chunk: list,
    file_path: Path,
    blog_config: BlogConfig,
    page_title: str,
):
    # TODO: Add templates from override
    # templates = jinja2.Environment(loader=jinja2.FileSystemLoader("templates/"))
    # print(templates.list_templates())
    # template = templates.get_template("index.html")

    index_template = resources.read_template_file(template_file_name="posts-list.html")
    context = {
        "posts": single_posts_chunk,
        "site_url": str(blog_config.mkdocs_config.site_url),
        "config": blog_config.plugin_config,
        "translation": blog_config.translation,
    }
    template = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(index_template)
    markdown = template.render(context)

    # TODO: when using pub-meta key name should be taken from plugin config
    page_meta = {"title": page_title}
    # TODO: consider moving slugify configuration to mkdocs.yaml
    if file_path.name.startswith("index-0"):
        slug = blog_config.plugin_config.slug
    elif file_path.name.startswith("index"):
        slug = f"{blog_config.plugin_config.slug}/{file_path.stem.split('-')[-1]}"
    else:
        slug = slugify(text=page_title.split("-")[-1].strip())
    page_meta["slug"] = slug
    page_meta["publish"] = "true"
    if not blog_config.plugin_config.searchable_non_posts:
        page_meta["search"] = {"exclude": True}  # type: ignore

    with open(file_path, mode="w") as teasers_index:
        teasers_index.write(f"---\n{yaml.dump(page_meta)}\n---\n\n{markdown}")



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging

from mkdocs.config import config_options as option
from mkdocs.config.base import Config


class _DebuggerConsoleConfig(Config):
    enabled = option.Type(bool, default=True)
    # noinspection PyUnresolvedReferences,PyProtectedMember
    log_level = option.Choice(choices=logging._nameToLevel.keys(), default="INFO")
    show_code_link = option.Type(bool, default=False)
    show_logger_name = option.Type(bool, default=True)
    show_entry_time = option.Type(bool, default=True)
    show_deprecation_warnings = option.Type(bool, default=False)
    entry_time_format = option.Type(str, default="%H:%M:%S.%f")
    filter_logger_names = option.Type(list, default=[])


class _DebuggerFileConfig(Config):
    enabled = option.Type(bool, default=True)
    # noinspection PyUnresolvedReferences,PyProtectedMember
    log_level = option.Choice(choices=logging._nameToLevel.keys(), default="DEBUG")
    log_format = option.Type(
        str, default="[%(created).14s][%(levelname)-5.5s][%(project_path)s:%(lineno)d] %(message)s"
    )
    remove_old_files = option.Type(bool, default=True)
    filter_logger_names = option.Type(list, default=[])


class _DebuggerZipConfig(Config):
    enabled = option.Type(bool, default=True)
    remove_old_files = option.Type(bool, default=True)
    add_pip_freeze = option.Type(bool, default=True)


class DebuggerConfig(Config):
    console_log: _DebuggerConsoleConfig = option.SubConfig(_DebuggerConsoleConfig)  # type: ignore
    file_log: _DebuggerFileConfig = option.SubConfig(_DebuggerFileConfig)  # type: ignore
    zip_log: _DebuggerZipConfig = option.SubConfig(_DebuggerZipConfig)  # type: ignore



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import importlib
import logging
import pathlib

import click


@click.group
@click.option("--debug", is_flag=True, default=False, help="Show debug messages")
def app(debug):
    """Publisher plugin for MkDocs (by Maciej 'maQ' Kusz).

    More information can be fount at: https://github.com/mkusz/mkdocs-publisher
    """
    log_level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=log_level)


def build_app(app: click.group):  # type: ignore
    plugins_path = pathlib.Path(__file__).parent.resolve() / "plugins"
    for plugin_file_name in pathlib.Path(plugins_path).rglob("*.py"):
        plugin_name = pathlib.PurePath(plugin_file_name).stem
        plugin_module = importlib.import_module(f"mkdocs_publisher._cli.plugins.{plugin_name}")
        app.add_command(
            cmd=getattr(plugin_module, "app", plugin_name), name=str(plugin_name.replace("_", "-"))
        )


build_app(app=app)


if __name__ == "__main__":
    app()



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from collections import OrderedDict
from copy import deepcopy
from typing import Optional

from mkdocs.structure.nav import Navigation
from mkdocs.structure.nav import Section
from mkdocs.structure.pages import Page

from mkdocs_publisher.blog.structures import BlogConfig

log = logging.getLogger("mkdocs.plugins.publisher.blog.modifiers")


def blog_post_nav_sorter(
    blog_config: BlogConfig,
    config_nav: OrderedDict,
):
    """Reorder blog posts in config navigation section from newest to oldest."""
    log.info("Reorder blog posts from newest to oldest")
    config_nav["_blog_posts_"] = [
        {blog_config.blog_posts[date].title: blog_config.blog_posts[date].path}
        for date in sorted(blog_config.blog_posts, reverse=True)
    ]


def blog_post_nav_remove(
    start_page: bool,
    blog_config: BlogConfig,
    nav: Navigation,
) -> None:
    """Remove blog posts pages, subindexes and section from direct navigation."""

    log.info("Removing blog posts pages and section from direct navigation")
    nav.items = [
        i for i in nav.items if not (isinstance(i, Section) and i.title.lower() == "_blog_posts_")
    ]
    log.info("Removing blog sub index pages from navigation menu")
    for item in nav.items:
        if (
            isinstance(item, Section)
            and item.title == blog_config.translation.blog_navigation_name
        ):
            children = []
            for section_item in item.children:
                if not (
                    isinstance(section_item, Page) and str(section_item.title).startswith("index-")
                ):
                    children.append(section_item)
                if (
                    isinstance(section_item, Page)
                    and not start_page
                    and str(section_item.title).startswith("index-0")
                ):
                    children.append(section_item)
            item.children = children


def blog_post_nav_next_prev_change(start_page: bool, blog_config: BlogConfig, page: Page):
    """Change blog post next/prev navigation"""

    if (start_page and page.title == "index") or (not start_page and page.title == "index-0"):
        page.title = blog_config.translation.recent_blog_posts_navigation_name  # type: ignore
        if page.next_page is not None and str(page.next_page.title).startswith("index-"):
            next_page_copy: Page = deepcopy(page.next_page)
            next_page_copy.title = blog_config.translation.older_posts  # type: ignore
            page.next_page = next_page_copy
    elif str(page.title).startswith("index") or (
        page.previous_page is not None and str(page.previous_page.title).startswith("index")
    ):
        page.title = blog_config.translation.older_posts  # type: ignore
        previous_page_copy: Optional[Page] = deepcopy(page.previous_page)
        previous_page_copy.title = blog_config.translation.newer_posts  # type: ignore
        page.previous_page = previous_page_copy
        if page.next_page is not None and str(page.next_page.title).startswith("index-"):
            next_page_copy: Page = deepcopy(page.next_page)
            next_page_copy.title = blog_config.translation.older_posts  # type: ignore
            page.next_page = next_page_copy



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from pathlib import Path
from typing import Optional
from typing import cast

from mkdocs.config.defaults import MkDocsConfig

# noinspection PyProtectedMember
from mkdocs_publisher._shared import links
from mkdocs_publisher._shared import mkdocs_utils

# noinspection PyProtectedMember
from mkdocs_publisher.blog.config import BlogPluginConfig
from mkdocs_publisher.obsidian.config import ObsidianPluginConfig

log = logging.getLogger("mkdocs.plugins.publisher.obsidian.md_links")


class MarkdownLinks:
    def __init__(self, mkdocs_config: MkDocsConfig):
        self._current_file_path: Optional[str] = None
        self._mkdocs_config: MkDocsConfig = mkdocs_config
        self._links_config: ObsidianPluginConfig = cast(
            ObsidianPluginConfig,
            mkdocs_utils.get_plugin_config(
                mkdocs_config=mkdocs_config, plugin_name="pub-obsidian"
            ),
        )
        self._blog_config: Optional[BlogPluginConfig] = cast(
            BlogPluginConfig,
            mkdocs_utils.get_plugin_config(mkdocs_config=mkdocs_config, plugin_name="pub-blog"),
        )

    @staticmethod
    def _normalize_wiki_embed_link(match: re.Match) -> str:
        wiki_embed_link = str(links.WikiEmbedLinkMatch(**match.groupdict()))
        log.debug(f"Normalizing wiki embed link: {match.group(0)} > {wiki_embed_link}")
        return wiki_embed_link

    @staticmethod
    def _normalize_wiki_link(match: re.Match) -> str:
        wiki_link_obj = links.LinkMatch(**match.groupdict())
        wiki_link_obj.is_wiki = True
        wiki_link = str(wiki_link_obj)
        log.debug(f"Normalizing wiki link: {match.group(0)} > {wiki_link}")
        return wiki_link

    def _normalize_md_embed_link(self, match: re.Match) -> str:
        md_embed_link_obj = links.MdEmbedLinkMatch(**match.groupdict())
        md_embed_link_obj.is_loading_lazy = self._links_config.links.img_lazy_loading
        md_embed_link = str(md_embed_link_obj)
        log.debug(f"Normalizing md embed link: {match.group(0)} > {md_embed_link}")
        return md_embed_link

    @staticmethod
    def _normalize_md_links(match: re.Match) -> str:
        md_link_obj = links.LinkMatch(**match.groupdict())
        md_link = str(md_link_obj)
        log.debug(f"Normalizing md link: {match.group(0)} > {md_link}")
        return md_link

    @staticmethod
    def _normalize_anchor_links(match: re.Match) -> str:
        anchor_link_obj = links.LinkMatch(**match.groupdict())
        anchor_link = str(anchor_link_obj)
        log.debug(f"Normalizing anchor link: {match.group(0)} > {anchor_link}")
        return anchor_link

    def normalize_links(self, markdown: str, current_file_path: str) -> str:
        self._current_file_path = current_file_path
        if self._links_config is not None and self._links_config.links.wikilinks_enabled:
            markdown = re.sub(links.WIKI_LINK_RE, self._normalize_wiki_link, markdown)
            markdown = re.sub(links.WIKI_EMBED_LINK_RE, self._normalize_wiki_embed_link, markdown)
            markdown = re.sub(links.ANCHOR_LINK_RE, self._normalize_anchor_links, markdown)
        markdown = re.sub(links.MD_EMBED_LINK_RE, self._normalize_md_embed_link, markdown)
        markdown = re.sub(links.MD_LINK_RE, self._normalize_md_links, markdown)
        return markdown

    def _normalize_relative_link(self, match: re.Match) -> str:
        md_link_obj = links.RelativeLinkMatch(**match.groupdict())
        md_link_obj.relative_path_finder = links.RelativePathFinder(
            current_file_path=Path(cast(str, self._current_file_path)),
            docs_dir=Path(self._mkdocs_config.docs_dir),
            relative_path=Path(cast(str, self._blog_config.blog_dir)),
        )
        return str(md_link_obj)

    def normalize_relative_links(self, markdown: str, current_file_path: str) -> str:
        self._current_file_path = current_file_path
        markdown = re.sub(links.RELATIVE_LINK_RE, self._normalize_relative_link, markdown)
        return markdown



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import urllib.parse
from typing import Optional
from typing import Union

import pymdownx.slugs

from mkdocs_publisher.meta.config import SlugModeChoiceEnum

log = logging.getLogger("mkdocs.plugins.publisher._shared.links")


def slugify(text: str) -> str:
    """Text slugify function that produces the same slug as MkDocs one"""

    text = urllib.parse.unquote(text)
    text = pymdownx.slugs.slugify(case="lower", normalize="NFD")(text=text, sep="-")
    return str(text).encode("ASCII", "ignore").decode("utf-8")


def create_slug(
    file_name: str,
    slug_mode: Union[SlugModeChoiceEnum, str],
    slug: Optional[str],
    title: Optional[str],
    warn_on_missing: bool = True,
) -> Optional[str]:
    """Generate slug for various modes"""
    # When slug from meta key is not present, try to get slug value the other way
    if slug is None or slug == "none":
        if slug_mode == SlugModeChoiceEnum.TITLE and title is not None:
            slug = slugify(text=title)
        elif slug_mode == SlugModeChoiceEnum.FILENAME:
            slug = slugify(text=file_name)

    # Log slug value
    if slug is None or slug == "none":
        slug = slugify(text=file_name)

        if warn_on_missing:
            log.warning(
                f'No slug for file "{file_name}" ' f"(mode: {slug_mode}). Fallback to file name."
            )

    log.debug(f'Slug for file "{file_name}" is: "{slug}"')
    return slug



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Callable
from typing import Optional

import jinja2
import watchdog.events
from mkdocs.config import Config
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.livereload import LiveReloadServer
from mkdocs.plugins import BasePlugin
from mkdocs.plugins import event_priority
from mkdocs.structure.files import Files
from mkdocs.structure.nav import Navigation
from mkdocs.structure.pages import Page
from mkdocs.utils import meta as meta_parser

# noinspection PyProtectedMember
from mkdocs_publisher._shared import resources

# noinspection PyProtectedMember
from mkdocs_publisher._shared.html_modifiers import HTMLModifier
from mkdocs_publisher.obsidian.backlinks import BacklinkLinks
from mkdocs_publisher.obsidian.backlinks import Link
from mkdocs_publisher.obsidian.callouts import CalloutToAdmonition
from mkdocs_publisher.obsidian.config import ObsidianPluginConfig
from mkdocs_publisher.obsidian.md_links import MarkdownLinks
from mkdocs_publisher.obsidian.vega import VegaCharts

COMMENTS_RE = re.compile(r"%%(?P<comment>[^%%]+)%%")

log = logging.getLogger("mkdocs.plugins.publisher.obsidian.plugin")


@dataclass
class Comment:
    comment: str
    is_html_comment: Optional[bool] = False

    def __repr__(self):
        return f"<!--{self.comment}-->" if self.is_html_comment else ""


class ObsidianPlugin(BasePlugin[ObsidianPluginConfig]):
    def __init__(self):
        self._backlink_links: Optional[BacklinkLinks] = None
        self._backlinks: dict[str, list[Link]] = {}
        self._md_links: Optional[MarkdownLinks] = None
        self._vega_pages: list[Page] = list()

    def _normalize_comments(self, match: re.Match) -> str:
        comment = Comment(**match.groupdict())
        comment.is_html_comment = self.config.comments.inject_as_html
        return str(comment)

    def on_config(self, config: MkDocsConfig) -> Optional[Config]:
        self._backlink_links = BacklinkLinks(mkdocs_config=config, backlinks=self._backlinks)
        self._md_links = MarkdownLinks(mkdocs_config=config)
        return config

    def on_files(self, files: Files, *, config: MkDocsConfig) -> Optional[Files]:
        if self.config.vega.enabled:
            resources.add_extra_css(
                stylesheet_file_name="obsidian.min.css",
                config=config,
                files=files,
            )
        return files

    def on_nav(
        self, nav: Navigation, *, config: MkDocsConfig, files: Files
    ) -> Optional[Navigation]:
        if self.config.backlinks.enabled:
            log.info("Parsing backlinks")
            for file in files:
                if file.page is not None:
                    log.debug(f"Parsing backlinks in file '{file.src_path}'")
                    with open(file.abs_src_path, encoding="utf-8-sig", errors="strict") as md_file:
                        markdown, _ = meta_parser.get_data(md_file.read())
                        markdown = self._md_links.normalize_links(
                            markdown=markdown, current_file_path=str(file.src_uri)
                        )
                        self._backlink_links.find_markdown_links(markdown=markdown, page=file.page)
        return nav

    @event_priority(100)  # Run before all other plugins
    def on_page_markdown(
        self, markdown: str, *, page: Page, config: MkDocsConfig, files: Files
    ) -> Optional[str]:
        # TODO: add verification if relative backlinks are enabled in .obsidian config
        markdown = self._md_links.normalize_links(
            markdown=markdown, current_file_path=str(page.file.src_uri)
        )
        if self.config.comments.enabled:
            markdown = re.sub(COMMENTS_RE, self._normalize_comments, markdown)

        if self.config.callouts.enabled:
            # TODO: add verification if all things are enabled in mkdocs.yaml config file
            callout_to_admonition = CalloutToAdmonition(callouts_config=self.config.callouts)
            markdown = callout_to_admonition.convert_callouts(
                markdown=markdown, file_path=str(page.file.src_uri)
            )

        if self.config.vega.enabled:
            # TODO: add verification if all things are enabled in .obsidian config
            vega_charts = VegaCharts(vega_config=self.config.vega)
            markdown = vega_charts.generate_charts(markdown=markdown)
            if vega_charts.is_vega:
                self._vega_pages.append(page)

        if self.config.backlinks.enabled:
            markdown = self._backlink_links.convert_to_backlink(markdown=markdown)
            page_backlinks = self._backlinks.get(f"{page.file.src_uri}", None)
            if page_backlinks is not None:
                log.debug(f"Adding backlinks to '{page.file.src_uri}'")
                backlink_template = resources.read_template_file(
                    template_file_name="backlinks.html"
                )
                context = {
                    "backlinks": page_backlinks,
                    "title": "Backlinks",  # TODO: move to translations
                }
                template = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(
                    backlink_template
                )
                markdown = f"{markdown}{template.render(context)}"
        return markdown

    def on_post_page(self, output: str, *, page: Page, config: MkDocsConfig) -> Optional[str]:
        if self.config.vega.enabled and page in self._vega_pages:
            # TODO: embed scripts to assets and give possibility to serve from site_dir
            html_modifier = HTMLModifier(markup=output)
            html_modifier.add_head_script(src="https://cdn.jsdelivr.net/npm/vega@5.22.1")
            html_modifier.add_head_script(src="https://cdn.jsdelivr.net/npm/vega-lite@5.6.1")
            html_modifier.add_head_script(src="https://cdn.jsdelivr.net/npm/vega-embed@6.21.2")
            output = str(html_modifier)

        return output

    # noinspection PyProtectedMember
    def on_serve(
        self, server: LiveReloadServer, *, config: MkDocsConfig, builder: Callable
    ) -> Optional[LiveReloadServer]:
        server.unwatch(config.docs_dir)

        docs_dirs_to_skip = [str(Path(config.docs_dir) / self.config.obsidian_dir)]

        # noinspection PyProtectedMember
        def no_obsidian_callback(event):
            """Watcher implementation that skips .obsidian directory"""
            if (
                isinstance(event, watchdog.events.FileModifiedEvent)
                and any([str(event.src_path).startswith(d) for d in docs_dirs_to_skip])
                or event.is_directory
            ):
                return

            with server._rebuild_cond:
                # noinspection PyProtectedMember
                server._to_rebuild[server.builder] = True
                server._rebuild_cond.notify_all()

        handler = watchdog.events.FileSystemEventHandler()
        handler.on_any_event = no_obsidian_callback

        if config.docs_dir in server._watched_paths:
            server._watched_paths[config.docs_dir] += 1
            return
        server._watched_paths[config.docs_dir] = 1

        server._watch_refs[config.docs_dir] = server.observer.schedule(
            handler, config.docs_dir, recursive=True
        )
        log.debug(f"Watching '{config.docs_dir}'")
        return server



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from pathlib import Path
from typing import Optional
from typing import Union
from typing import cast

from mkdocs.config.defaults import MkDocsConfig

from mkdocs_publisher._shared import mkdocs_utils
from mkdocs_publisher.blog.config import BlogPluginConfig
from mkdocs_publisher.obsidian.config import ObsidianPluginConfig

log = logging.getLogger("mkdocs.plugins.publisher._shared.publisher_utils")


def get_blog_dir(mkdocs_config: MkDocsConfig) -> Optional[Path]:
    blog_config: Optional[Union[BlogPluginConfig, Path]] = cast(
        BlogPluginConfig,
        mkdocs_utils.get_plugin_config(
            mkdocs_config=mkdocs_config,
            plugin_name="pub-blog",
        ),
    )
    if blog_config is not None:
        blog_config = Path(mkdocs_config.docs_dir).joinpath(blog_config.blog_dir)

    return blog_config


def get_obsidian_dirs(mkdocs_config: MkDocsConfig) -> tuple[list[Path], Optional[Path]]:
    ignored_dirs: list[Path] = []
    attachments_dir: Optional[Path] = None
    docs_dir = Path(mkdocs_config.docs_dir)

    obsidian_config: Optional[ObsidianPluginConfig] = cast(
        ObsidianPluginConfig,
        mkdocs_utils.get_plugin_config(
            mkdocs_config=mkdocs_config,
            plugin_name="pub-obsidian",
        ),
    )
    if obsidian_config is not None:
        ignored_dirs.append(docs_dir.joinpath(obsidian_config.obsidian_dir))
        ignored_dirs.append(docs_dir.joinpath(obsidian_config.templates_dir))
        attachments_dir = docs_dir.joinpath(obsidian_config.attachments_dir)

    return ignored_dirs, attachments_dir



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
import site
from datetime import datetime
from pathlib import Path

import colorama

from mkdocs_publisher.debugger.config import _DebuggerConsoleConfig
from mkdocs_publisher.debugger.config import _DebuggerFileConfig

colorama.init()

LIVERELOAD_MSG_RE = re.compile(r"(?P<time>\[\d{,2}:\d{1,2}:\d{,2}] )?(?P<text>.*)")
DEPRECATION_MSG_RE = re.compile(r"^(?P<deprecation>DeprecationWarning:)")
SITE_PACKAGES_DIR = Path(site.getsitepackages()[0])
LOG_LEVEL_COLOR_MAPPING = {
    logging.DEBUG: colorama.Fore.BLUE,
    logging.INFO: colorama.Fore.GREEN,
    logging.WARNING: colorama.Fore.YELLOW,
    logging.ERROR: colorama.Fore.RED,
    logging.CRITICAL: colorama.Fore.RED,
}


class DatedFileHandler(logging.FileHandler):
    def __init__(self, filename):
        dated_filename = datetime.utcnow().strftime(str(filename))
        Path(dated_filename).parent.mkdir(parents=True, exist_ok=True)
        super().__init__(filename=dated_filename)


class ProjectPathStreamFormatter(logging.Formatter):
    def __init__(self, console_config: _DebuggerConsoleConfig):
        self._console_config: _DebuggerConsoleConfig = console_config

        super().__init__()

    @staticmethod
    def _livereload_msg_strip_time(match: re.Match) -> str:
        return match.groupdict()["text"]

    def format(self, record: logging.LogRecord) -> str:
        path = Path(record.pathname)
        try:
            record.project_path = str(path.relative_to(path.cwd()))
        except ValueError:
            record.project_path = str(path)

        fmt = "%(project_path)s:%(lineno)d " if self._console_config.show_code_link else ""

        if self._console_config.show_entry_time:
            fmt = f"{fmt}{colorama.Fore.MAGENTA}%(asctime)s{colorama.Fore.RESET} "

        fmt = (
            f'{fmt}[{LOG_LEVEL_COLOR_MAPPING.get(record.levelno) or ""}'
            f"%(levelname)-5.5s{colorama.Fore.RESET}] %(message)s"
        )

        if self._console_config.show_logger_name:
            fmt = f"{fmt} {colorama.Fore.CYAN}[%(name)s]{colorama.Fore.RESET}"

        self._style._fmt = fmt
        self.datefmt = str(self._console_config.entry_time_format).replace(
            "%f", str(record.msecs)[0:3]
        )

        if self._console_config.show_entry_time:
            record.msg = re.sub(LIVERELOAD_MSG_RE, self._livereload_msg_strip_time, record.msg)

        return super().format(record=record)


class ProjectPathFileFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        project_file_path = Path(record.pathname)
        try:
            record.project_path = str(
                project_file_path.relative_to(
                    SITE_PACKAGES_DIR
                    if "site-packages" in str(project_file_path)
                    else project_file_path.cwd()
                )
            )
        except ValueError:
            record.project_path = str(project_file_path)

        return super().format(record=record)


class ProjectPathConsoleFilter(logging.Filter):
    def __init__(self, console_config: _DebuggerConsoleConfig):
        self._console_config: _DebuggerConsoleConfig = console_config
        super().__init__()

    def filter(self, record):
        if not self._console_config.show_deprecation_warnings and re.findall(
            DEPRECATION_MSG_RE, record.msg
        ):
            return None
        return record if record.name not in self._console_config.filter_logger_names else None


class ProjectPathFileFilter(logging.Filter):
    def __init__(self, file_config: _DebuggerFileConfig):
        self._file_config: _DebuggerFileConfig = file_config
        super().__init__()

    def filter(self, record):
        return record if record.name not in self._file_config.filter_logger_names else None



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from pathlib import Path
from typing import Optional

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.nav import Link
from mkdocs.structure.nav import Section
from mkdocs.structure.pages import Page

from mkdocs_publisher.meta.meta_files import MetaFile
from mkdocs_publisher.meta.meta_files import MetaFiles

log = logging.getLogger("mkdocs.plugins.publisher.meta.nav")


class MetaNav:
    def __init__(self, meta_files: MetaFiles, blog_dir: Optional[Path] = None):
        self._meta_files: MetaFiles = meta_files
        self._blog_dir: Optional[Path] = blog_dir

    def nav_cleanup(self, items, removal_list: list[str]) -> list:
        nav = []
        for item in items:
            if isinstance(item, Section):
                item.children = self.nav_cleanup(items=item.children, removal_list=removal_list)
                # If section is empty, skip it
                if len(item.children) > 0:
                    nav.append(item)
            elif (isinstance(item, Page) and str(item.file.src_path) not in removal_list) or (
                isinstance(item, Link) and item.title not in removal_list
            ):
                nav.append(item)
        return nav

    def _build_nav(self, meta_files_gen, current_dir: Path) -> tuple[list, Optional[MetaFile]]:
        nav = []
        meta_file: Optional[MetaFile] = None
        while True:
            if meta_file is None:
                try:
                    meta_file = next(meta_files_gen)
                except StopIteration:
                    break
            # Iterate over meta links until last one
            if meta_file.is_dir and meta_file.abs_path.is_relative_to(current_dir):
                overview_path: Path = meta_file.abs_path.joinpath(self._meta_files.meta_file)
                overview_nav: list[Path] = (
                    [meta_file.path.joinpath(self._meta_files.meta_file)]
                    if meta_file.is_overview and overview_path.exists()
                    else []
                )

                title = meta_file.title
                prev_path = meta_file.path
                sub_nav, meta_file = self._build_nav(
                    meta_files_gen=meta_files_gen, current_dir=meta_file.abs_path
                )
                sub_nav = [*overview_nav, *sub_nav]
                if sub_nav:
                    if prev_path == self._blog_dir:
                        nav.append({str(prev_path): str(prev_path)})
                    else:
                        nav.append({title: sub_nav})
            elif meta_file.is_dir:
                return nav, meta_file  # Jump to subdirectory
            elif (
                not meta_file.is_dir and not meta_file.is_draft and meta_file.path.suffix == ".md"
            ):
                if meta_file.abs_path.is_relative_to(current_dir):
                    nav.append({meta_file.title: str(meta_file.path)})
                    meta_file = None  # File added, skip to next
                else:
                    return nav, meta_file  # Jump to subdirectory
            else:
                meta_file = None  # File not to be processed, skip to next
        return nav, None

    def build_nav(self, mkdocs_config: MkDocsConfig) -> list:
        nav, _ = self._build_nav(
            meta_files_gen=self._meta_files.files_gen(),
            current_dir=Path(mkdocs_config.docs_dir),
        )
        return nav



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import subprocess
from hashlib import md5
from pathlib import Path
from typing import Optional
from uuid import uuid4

log = logging.getLogger("mkdocs.plugins.publisher._shared.file_utils")


def run_subprocess(cmd, capture_output: bool = True) -> subprocess.CompletedProcess:
    cmd = [arg for arg in cmd if arg is not None]
    log.debug(f"Run cmd: {' '.join(cmd)}")
    return subprocess.run(cmd, capture_output=capture_output)


def remove_dir(directory: Path):
    for file in directory.iterdir():
        if file.is_dir():
            remove_dir(directory=file)
            file.rmdir()
        else:
            file.unlink(missing_ok=True)


def calculate_file_hash(file: Path, block_size: int = 8192) -> Optional[str]:
    try:
        with open(file, "rb") as binary_file:
            file_hash = md5()
            while chunk := binary_file.read(block_size):
                file_hash.update(chunk)
            return file_hash.hexdigest()
    except IsADirectoryError:
        return None


def list_files(
    directory: Path,
    extensions: Optional[list[str]] = None,
    exclude: Optional[list[str]] = None,
) -> list[Path]:
    temp_files_list: list[Path] = []
    extensions = [] if extensions is None else extensions
    exclude = [] if exclude is None else exclude
    for ext in extensions:
        for file in directory.glob(f"**/*{ext}"):
            temp_files_list.append(file.relative_to(directory))

    excluded_files_list: list[Path] = []
    for exc in exclude:
        for file in directory.rglob(exc):
            excluded_files_list.append(file.relative_to(directory))

    files_list: list[Path] = [file for file in temp_files_list if file not in excluded_files_list]
    return files_list


def get_hashed_file_name(file: Path) -> Path:
    return Path(file.with_name(f"{str(uuid4()).replace('-', '')}{str(file.suffix).lower()}").name)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mkdocs.config import config_options as option
from mkdocs.config.base import Config


class _ObsidianBacklinksConfig(Config):
    enabled = option.Type(bool, default=True)


class _ObsidianCalloutsConfig(Config):
    enabled = option.Type(bool, default=True)
    indentation = option.Choice(["tabs", "spaces"], default="spaces")


class _ObsidianCommentsConfig(Config):
    enabled = option.Type(bool, default=True)
    inject_as_html = option.Type(bool, default=False)


class _ObsidianVegaConfig(Config):
    enabled = option.Type(bool, default=True)
    vega_schema = option.Type(str, default="https://vega.github.io/schema/vega/v5.json")
    vega_lite_schema = option.Type(str, default="https://vega.github.io/schema/vega-lite/v5.json")


class _ObsidianLinksConfig(Config):
    wikilinks_enabled = option.Type(bool, default=True)
    img_lazy_loading = option.Type(bool, default=True)


class ObsidianPluginConfig(Config):
    # TODO: read those values from Obsidian config files
    obsidian_dir = option.Type(str, default=".obsidian")
    templates_dir = option.Type(str, default="_templates")
    attachments_dir = option.Type(str, default="_attachments")

    backlinks: _ObsidianBacklinksConfig = option.SubConfig(_ObsidianBacklinksConfig)  # type: ignore
    callouts: _ObsidianCalloutsConfig = option.SubConfig(_ObsidianCalloutsConfig)  # type: ignore
    comments: _ObsidianCommentsConfig = option.SubConfig(_ObsidianCommentsConfig)  # type: ignore
    vega: _ObsidianVegaConfig = option.SubConfig(_ObsidianVegaConfig)  # type: ignore
    links: _ObsidianLinksConfig = option.SubConfig(_ObsidianLinksConfig)  # type: ignore



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import contextlib
import logging
from datetime import datetime
from pathlib import Path
from typing import Literal
from typing import Optional

from mkdocs.config import Config
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.plugins import event_priority
from mkdocs.structure.files import Files
from mkdocs.structure.nav import Navigation
from mkdocs.structure.pages import Page

# noinspection PyProtectedMember
from mkdocs_publisher._shared import publisher_utils
from mkdocs_publisher._shared.html_modifiers import HTMLModifier
from mkdocs_publisher.meta.config import MetaPluginConfig
from mkdocs_publisher.meta.meta_files import MetaFiles
from mkdocs_publisher.meta.nav import MetaNav

log = logging.getLogger("mkdocs.plugins.publisher.meta.plugin")


class MetaPlugin(BasePlugin[MetaPluginConfig]):
    def __init__(self):
        """
        TODO: SEO optimizations
        - https://octamedia.pl/blog/linkowanie-wewnetrzne/ (useful for obsidian backlinks)
        """

        self._on_serve = False
        self._attachments_dir: Optional[Path] = None
        self._ignored_dirs: list[Path] = []
        self._meta_files: MetaFiles = MetaFiles()
        self._meta_nav: Optional[MetaNav] = None

    def on_startup(
        self, *, command: Literal["build", "gh-deploy", "serve"], dirty: bool
    ) -> None:  # pragma: no cover
        if command == "serve":
            self._on_serve = True
        self._meta_files.on_serve = self._on_serve

    @event_priority(100)  # Run before any other plugins
    def on_config(self, config: MkDocsConfig) -> Optional[Config]:
        # Set some default values
        log.info("Read files and directories metadata")
        blog_dir: Optional[Path] = publisher_utils.get_blog_dir(mkdocs_config=config)
        self._meta_nav = MetaNav(
            meta_files=self._meta_files,
            blog_dir=blog_dir.relative_to(config.docs_dir) if blog_dir else blog_dir,
        )
        self._ignored_dirs, self._attachments_dir = publisher_utils.get_obsidian_dirs(
            mkdocs_config=config
        )  # pragma: no cover
        self._meta_files.set_configs(
            mkdocs_config=config, meta_plugin_config=self.config
        )  # pragma: no cover
        self._meta_files.add_hidden_path(hidden_path=self._attachments_dir)  # pragma: no cover
        self._meta_files.add_meta_files(ignored_dirs=self._ignored_dirs)  # pragma: no cover

        log.info(
            f"Ignored directories: "
            f"{[str(d.relative_to(config.docs_dir)) for d in self._ignored_dirs]}"
        )
        log.info(f"Draft files and directories: " f"{list(self._meta_files.drafts().keys())}")
        log.info(f"Hidden files and directories: " f"{list(self._meta_files.hidden().keys())}")

        config.nav = self._meta_nav.build_nav(mkdocs_config=config)

        return config

    @event_priority(-100)
    def on_files(
        self, files: Files, *, config: MkDocsConfig
    ) -> Optional[Files]:  # pragma: no cover
        new_files = self._meta_files.change_files_slug(
            files=files, ignored_dirs=self._ignored_dirs
        )

        return new_files

    def on_nav(
        self, nav: Navigation, *, config: MkDocsConfig, files: Files
    ) -> Optional[Navigation]:  # pragma: no cover
        removal_list = [*self._meta_files.drafts().keys(), *self._meta_files.hidden().keys()]

        log.debug(f"Nav elements to remove: {removal_list}")
        nav.items = self._meta_nav.nav_cleanup(
            items=nav.items,
            removal_list=removal_list,
        )
        return nav

    @event_priority(-100)  # Run after all other plugins
    def on_page_markdown(self, markdown: str, *, page: Page, config: MkDocsConfig, files: Files):
        # Modify page update date
        # TODO: move date format to config
        # TODO: warn on missing in config
        update_date: datetime = page.meta.get(
            "update", page.meta.get("date", datetime.strptime(page.update_date, "%Y-%m-%d"))
        )
        with contextlib.suppress(AttributeError):
            page.update_date = update_date.strftime("%Y-%m-%d")

        # Conditionally exclude file from Material for MkDocs search plugin
        if (
            page.file.src_uri in self._meta_files.drafts(files=True)
            and not self.config.publish.search_in_draft
        ) or (
            page.file.src_uri in self._meta_files.hidden(files=True)
            and not self.config.publish.search_in_hidden
        ):
            page.meta["search"] = {"exclude": True}

    @event_priority(-100)  # Run after all other plugins
    def on_post_page(self, output: str, *, page: Page, config: MkDocsConfig) -> Optional[str]:
        html_modifier = HTMLModifier(markup=output)
        html_modifier.fix_img_links()

        return str(html_modifier)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from collections import OrderedDict
from dataclasses import fields
from datetime import datetime
from pathlib import Path

from mkdocs.utils import meta as meta_parser

from mkdocs_publisher.blog.structures import BlogConfig
from mkdocs_publisher.blog.structures import BlogPost

log = logging.getLogger("mkdocs.plugins.publisher.blog.parsers")

REQUIRED_META_KEYS = ["title", "date", "slug", "tags", "categories", "description", "publish"]
# TODO: read it from pub-meta if configured


def count_words(content):
    """Count words in markdown content.

    This code is based on: https://github.com/gandreadis/markdown-word-count
    """
    content = re.sub(r"<!--(.*?)-->", "", content, flags=re.MULTILINE)  # Comments
    content = content.replace("\t", "    ")  # Tabs to spaces
    content = re.sub(r"[ ]{2,}", "    ", content)  # More than 1 space to 4 spaces
    content = re.sub(r"^\[[^]]*][^(].*", "", content, flags=re.MULTILINE)  # Footnotes
    content = re.sub(
        r"^( {4,}[^-*]).*", "", content, flags=re.MULTILINE
    )  # Indented blocks of code
    content = re.sub(r"{#.*}", "", content)  # Custom header IDs
    content = content.replace("\n", " ")  # Replace newlines with spaces for uniform handling
    content = re.sub(r"!\[[^]]*]\([^)]*\)", "", content)  # Remove images
    content = re.sub(r"</?[^>]*>", "", content)  # Remove HTML tags
    content = re.sub(r"[#*`~\-–^=<>+|/:]", "", content)  # Remove special characters
    content = re.sub(r"\[[0-9]*]", "", content)  # Remove footnote references
    content = re.sub(r"[0-9#]*\.", "", content)  # Remove enumerations

    return len(content.split())


# from mkdocs.utils import meta as meta_parser
# with file.open(encoding="utf-8-sig", errors="strict") as md_file:
#     markdown, meta = meta_parser.get_data(md_file.read())


def parse_markdown_files(
    blog_config: BlogConfig,
    config_nav: OrderedDict,
    on_serve: bool = False,
):
    """Parse all markdown files and extract blog posts from `blog_dir` (default: 'posts').
    BlogPost object is created and filled with content and metadata of the post.
    """
    log.info(f"Parsing blog posts from '{blog_config.blog_dir}' directory")
    for file_path in blog_config.docs_dir.glob("**/*.md"):
        if (
            blog_config.meta_config is not None
            and file_path.parts[-1] == blog_config.meta_config.dir_meta_file
        ):
            continue
        file_path = Path(file_path)
        path = Path(file_path).relative_to(blog_config.docs_dir)
        if path.is_relative_to(blog_config.blog_dir):
            parents = list(path.parents)[:-1]
            with file_path.open(encoding="utf-8-sig", errors="strict") as md_file:
                content, post_meta = meta_parser.get_data(md_file.read())
                if not parents:
                    for line in content.split("\n"):
                        if line.startswith("# "):
                            config_nav[line[2:]] = str(path)
                elif str(parents[0]) == str(blog_config.blog_dir):
                    if "publish" not in post_meta:
                        # TODO: read default value from meta config
                        log.info(
                            f"File: {file_path} - missing 1 required positional argument: "
                            f"'publish' (setting to default: draft)"
                        )
                        post_meta["publish"] = "draft"

                    # Skip non-published
                    if not on_serve and post_meta["publish"] not in ["published", "true", True]:
                        # TODO: make it configurable
                        continue

                    # Convert tags format
                    if "tags" in post_meta and post_meta["tags"] is not None:
                        if "," in post_meta["tags"]:
                            # noinspection PyUnresolvedReferences
                            post_meta["tags"] = [t.strip() for t in post_meta["tags"].split(",")]
                        elif isinstance(post_meta["tags"], str):
                            post_meta["tags"] = [post_meta["tags"]]
                    else:
                        post_meta["tags"] = ["undefined"]  # TODO: move this value to config

                    # Convert categories format
                    if "categories" in post_meta and post_meta["categories"] is not None:
                        if "," in post_meta["categories"]:
                            post_meta["categories"] = [
                                t.strip() for t in post_meta["categories"].split(",")
                            ]
                        elif isinstance(post_meta["categories"], str):
                            post_meta["categories"] = [post_meta["categories"]]
                    else:
                        post_meta["categories"] = ["undefined"]  # TODO: move this value to config

                    # Setup date to current one, when missing
                    if "date" in post_meta and post_meta["date"] is None:
                        post_meta["date"] = datetime.utcnow()

                    # Create a new blog post
                    blog_post_keys = [f.name for f in fields(BlogPost)]
                    post_data = {k: v for k, v in post_meta.items() if k in blog_post_keys}
                    post_data["content"] = content
                    post_data["path"] = str(path)
                    if (
                        "slug" in post_meta
                        and post_meta["slug"] is not None
                        and post_meta["slug"].strip() != ""
                    ):
                        post_data["slug"] = post_meta["slug"]

                    blog_post: BlogPost = BlogPost(**post_data)

                    # Add new post to blog posts collection
                    blog_config.blog_posts[blog_post.date] = blog_post
                    log.debug(f"New blog posts: {blog_post.title}")

                    # TODO: add reading time
                    # print(f"{file_path} - {count_words(post.content) / 265 * 60}")


def create_blog_post_teaser(blog_config: BlogConfig):
    """Extracting beginning of a blog content as a teaser. End of a teaser is determined by
    'teaser_marker' config value (default: '<!-- more -->'). When not found entire blog post
    is considered a teaser and link to full blog posts is not displayed, but whole blog post
    is still rendered, so it can be navigated to from other blog post using next/previous page
    arrows at the bottom of the page.
    """
    for date in sorted(blog_config.blog_posts, reverse=True):
        content = []
        for line in blog_config.blog_posts[date].content.split("\n"):  # type: ignore
            if line == blog_config.plugin_config.teaser_marker:
                blog_config.blog_posts[date].is_teaser = True
                blog_config.blog_posts[date].teaser = "\n".join(c for c in content)
            content.append(line)
        if not blog_config.blog_posts[date].is_teaser:
            blog_config.blog_posts[date].teaser = "\n".join(c for c in content)
        log.debug(
            f"Post: '{blog_config.blog_posts[date].title}' "
            f"is a teaser: {blog_config.blog_posts[date].is_teaser}"
        )
